#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from preggy import expect
from tornado.testing import gen_test

from tests.fixtures.images import default_image
from tests.handlers.test_base_handler import BaseImagingTestCase
from thumbor.config import Config
from thumbor.context import Context, ServerParameters
from thumbor.importer import Importer


class ImageOperationsWithStoredKeysTestCase(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(SECURITY_KEY="MYKEY")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path
        cfg.ALLOW_UNSAFE_URL = False
        cfg.STORES_CRYPTO_KEY_FOR_EACH_IMAGE = True

        cfg.STORAGE = "thumbor.storages.file_storage"
        cfg.STORAGE_FILE_STORAGE_ROOT_PATH = self.root_path

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8891, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "MYKEY"
        return Context(server, cfg, importer)

    @gen_test
    async def test_stored_security_key_with_regular_image(self):
        storage = self.context.modules.storage
        self.context.server.security_key = "MYKEY"
        await storage.put_crypto(
            "image.jpg"
        )  # Write a file on the file storage containing the security key
        self.context.server.security_key = "MYKEY2"

        try:
            response = await self.async_fetch(
                "/nty7gpBIRJ3GWtYDLLw6q1PgqTo=/smart/image.jpg"
            )
            expect(response.code).to_equal(200)
            expect(response.body).to_be_similar_to(default_image())
        finally:
            self.context.server.security_key = "MYKEY"

    @gen_test
    async def test_stored_security_key_with_regular_image_with_querystring(
        self,
    ):  # NOQA
        storage = self.context.modules.storage
        self.context.server.security_key = "MYKEY"
        await storage.put_crypto(
            "image.jpg%3Fts%3D1"
        )  # Write a file on the file storage containing the security key
        self.context.server.security_key = "MYKEY2"

        response = await self.async_fetch(
            "/Iw7LZGdr-hHj2gQ4ZzksP3llQHY=/smart/image.jpg%3Fts%3D1"
        )
        expect(response.code).to_equal(200)
        expect(response.body).to_be_similar_to(default_image())

    @gen_test
    async def test_stored_security_key_with_regular_image_with_hash(self):
        storage = self.context.modules.storage
        self.context.server.security_key = "MYKEY"
        await storage.put_crypto(
            "image.jpg%23something"
        )  # Write a file on the file storage containing the security key
        self.context.server.security_key = "MYKEY2"

        response = await self.async_fetch(
            "/fxOHtHcTZMyuAQ1YPKh9KWg7nO8=/smart/image.jpg%23something"
        )
        expect(response.code).to_equal(200)
        expect(response.body).to_be_similar_to(default_image())



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from unittest import mock

from preggy import expect

import thumbor.metrics
from tests.base import TestCase
from thumbor.importer import Importer


class DefaultMetricsTestCase(TestCase):
    def get_importer(self):
        importer = Importer(self.config)
        importer.import_modules()
        return importer

    def test_can_create_context_with_default_metrics(self):
        expect(self.context).not_to_be_null()
        expect(self.context.metrics).to_be_instance_of(
            thumbor.metrics.logger_metrics.Metrics
        )

    @mock.patch("thumbor.metrics.BaseMetrics.initialize")
    def test_can_initizalize_when_request_comes(self, mocked_initialize):
        expect(mocked_initialize.call_count).to_equal(0)
        self.fetch("/unsafe/smart/image.jpg")
        expect(mocked_initialize.call_count).to_equal(1)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com


class FocalPoint:
    ALIGNMENT_PERCENTAGES = {
        "left": 0.0,
        "center": 0.5,
        "right": 1.0,
        "top": 0.0,
        "middle": 0.5,
        "bottom": 1.0,
    }

    def to_dict(self):
        return {
            "x": self.x,
            "y": self.y,
            "z": self.weight,
            "height": self.height,
            "width": self.width,
            "origin": self.origin,
        }

    @classmethod
    def from_dict(cls, values):
        return cls(
            x=int(values["x"]),
            y=int(values["y"]),
            weight=int(values["z"]),
            width=int(values.get("width", 1)),
            height=int(values.get("height", 1)),
            origin=values.get("origin", "alignment"),
        )

    def __init__(
        self,
        x,  # pylint: disable=invalid-name
        y,  # pylint: disable=invalid-name
        height=1,
        width=1,
        weight=1.0,
        origin="alignment",
    ):
        self.x = int(x)  # pylint: disable=invalid-name
        self.y = int(y)  # pylint: disable=invalid-name
        self.height = int(height)
        self.width = int(width)
        self.weight = weight
        self.origin = origin

    @classmethod
    def from_square(
        cls, x, y, width, height, origin="detection"
    ):  # pylint: disable=invalid-name
        center_x = x + width // 2
        center_y = y + height // 2
        return cls(
            center_x,
            center_y,
            height=height,
            width=width,
            weight=width * height,
            origin=origin,
        )

    @classmethod
    def from_alignment(cls, halign, valign, width, height):
        x_coord = width * cls.ALIGNMENT_PERCENTAGES[halign]
        y_coord = height * cls.ALIGNMENT_PERCENTAGES[valign]

        return cls(x_coord, y_coord)

    def __repr__(self):
        return (
            f"FocalPoint(x: {self.x}, y: {self.y}, width: {self.width}, "
            f"height: {self.height}, weight: {self.weight:.0f}, origin: {self.origin})"
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from os.path import abspath
from unittest.mock import patch

from preggy import expect
from tornado.testing import gen_test

import thumbor.detectors
from tests.base import DetectorTestCase
from thumbor.detectors.face_detector import Detector as FaceDetector


class FaceDetectorTestCase(DetectorTestCase):
    @gen_test
    async def test_should_detect_one_face(self):
        with open(
            abspath(
                "./tests/fixtures/images/Giunchedi%2C_Filippo_January_2015_01.jpg"
            ),
            "rb",
        ) as fixture:
            self.engine.load(fixture.read(), None)

        for i, detector in enumerate(
            [
                "./thumbor/detectors/face_detector/"
                "haarcascade_frontalface_default.xml",
                "./thumbor/detectors/face_detector/haarcascade_frontalface_alt.xml",
                "./thumbor/detectors/face_detector/haarcascade_frontalface_alt2.xml",
                "./thumbor/detectors/face_detector/"
                "haarcascade_frontalface_alt_tree.xml",
            ]
        ):
            self.context.config.FACE_DETECTOR_CASCADE_FILE = abspath(detector)

            if hasattr(FaceDetector, "cascade"):
                del FaceDetector.cascade
            await FaceDetector(self.context, 0, None).detect()
            detection_result = self.context.request.focal_points[i]
            expect(detection_result.origin).to_equal("Face Detection")
            expect(detection_result.x).to_be_numeric()
            expect(detection_result.y).to_be_numeric()
            expect(detection_result.width).to_be_numeric()
            expect(detection_result.height).to_be_numeric()

    @gen_test
    async def test_should_skip_if_opencv_not_found(self):
        with open(
            abspath(
                "./tests/fixtures/images/Giunchedi%2C_Filippo_January_2015_01.jpg"
            ),
            "rb",
        ) as fixture:
            # ARRANGE
            self.engine.load(fixture.read(), None)
        cascade = "./thumbor/detectors/face_detector/haarcascade_frontalface_default.xml"
        self.context.config.FACE_DETECTOR_CASCADE_FILE = abspath(cascade)

        with patch.object(
            thumbor.detectors.BaseDetector, "verify_cv", lambda self: False
        ):
            if hasattr(FaceDetector, "cascade"):
                del FaceDetector.cascade

            # ACT
            await FaceDetector(self.context, 0, None).detect()

            # ASSERT
            detection_result = self.context.request.focal_points
            expect(detection_result).to_length(0)

    @gen_test
    async def test_should_not_detect(self):
        with open(
            abspath("./tests/fixtures/images/no_face.jpg"), "rb"
        ) as fixture:
            self.engine.load(fixture.read(), None)

        self.context.config.FACE_DETECTOR_CASCADE_FILE = abspath(
            "./thumbor/detectors/face_detector/haarcascade_frontalface_default.xml"
        )

        await FaceDetector(self.context, 0, []).detect()
        expect(self.context.request.focal_points).to_be_empty()

    @gen_test
    async def test_should_run_on_grayscale_images(self):
        with open(
            abspath(
                "./tests/fixtures/images/"
                "Giunchedi%2C_Filippo_January_2015_01-grayscale.jpg"
            ),
            "rb",
        ) as fixture:
            self.engine.load(fixture.read(), None)

        self.context.config.FACE_DETECTOR_CASCADE_FILE = abspath(
            "./thumbor/detectors/face_detector/haarcascade_frontalface_default.xml",
        )

        if hasattr(FaceDetector, "cascade"):
            del FaceDetector.cascade
        await FaceDetector(self.context, 0, None).detect()
        detection_result = self.context.request.focal_points[0]
        expect(detection_result.origin).to_equal("Face Detection")
        expect(detection_result.x).to_be_numeric()
        expect(detection_result.y).to_be_numeric()
        expect(detection_result.width).to_be_numeric()
        expect(detection_result.height).to_be_numeric()

    @gen_test
    async def test_should_run_on_cmyk_images(self):
        with open(
            abspath(
                "./tests/fixtures/images/"
                "Giunchedi%2C_Filippo_January_2015_01-cmyk.jpg"
            ),
            "rb",
        ) as fixture:
            self.engine.load(fixture.read(), None)

        self.context.config.FACE_DETECTOR_CASCADE_FILE = abspath(
            "./thumbor/detectors/face_detector/haarcascade_frontalface_default.xml",
        )

        if hasattr(FaceDetector, "cascade"):
            del FaceDetector.cascade
        await FaceDetector(self.context, 0, None).detect()
        detection_result = self.context.request.focal_points[0]
        expect(detection_result.origin).to_equal("Face Detection")
        expect(detection_result.x).to_be_numeric()
        expect(detection_result.y).to_be_numeric()
        expect(detection_result.width).to_be_numeric()
        expect(detection_result.height).to_be_numeric()

    @gen_test
    async def test_should_run_on_images_with_alpha(self):
        with open(
            abspath(
                "./tests/fixtures/images/Giunchedi%2C_Filippo_January_2015_01.png"
            ),
            "rb",
        ) as fixture:
            self.engine.load(fixture.read(), None)

        self.context.config.FACE_DETECTOR_CASCADE_FILE = abspath(
            "./thumbor/detectors/face_detector/haarcascade_frontalface_default.xml",
        )

        if hasattr(FaceDetector, "cascade"):
            del FaceDetector.cascade
        await FaceDetector(self.context, 0, None).detect()
        detection_result = self.context.request.focal_points[0]
        expect(detection_result.origin).to_equal("Face Detection")
        expect(detection_result.x).to_be_numeric()
        expect(detection_result.y).to_be_numeric()
        expect(detection_result.width).to_be_numeric()
        expect(detection_result.height).to_be_numeric()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from json import loads
from shutil import which

from preggy import expect
from tornado.testing import gen_test

from tests.handlers.test_base_handler import BaseImagingTestCase
from thumbor.config import Config
from thumbor.context import Context, ServerParameters
from thumbor.engines.pil import Engine
from thumbor.importer import Importer
from thumbor.storages.file_storage import Storage as FileStorage
from thumbor.storages.no_storage import Storage as NoStorage

# pylint: disable=broad-except,abstract-method,attribute-defined-outside-init,line-too-long,too-many-public-methods
# pylint: disable=too-many-lines


class StorageOverrideTestCase(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path
        cfg.STORAGE = "thumbor.storages.file_storage"
        cfg.FILE_STORAGE_ROOT_PATH = self.root_path

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        return Context(server, cfg, importer)

    @gen_test
    async def test_shouldnt_call_put_when_storage_overridden_to_nostorage(
        self,
    ):  # NOQA
        old_load = Engine.load
        old_put = FileStorage.put

        def load_override(self, arg, arg2):
            self.context.modules.storage = NoStorage(None)
            return old_load(self, arg, arg2)

        def put_override(*_):
            expect.not_to_be_here()

        Engine.load = load_override
        FileStorage.put = put_override

        response = await self.async_fetch("/unsafe/image.jpg")

        Engine.load = old_load
        FileStorage.put = old_put

        expect(response.code).to_equal(200)


class ImageOperationsWithoutStorage(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path
        cfg.STORAGE = "thumbor.storages.no_storage"
        cfg.AUTO_WEBP = True
        cfg.USE_GIFSICLE_ENGINE = True
        cfg.RESPECT_ORIENTATION = True

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        ctx = Context(server, cfg, importer)
        ctx.server.gifsicle_path = which("gifsicle")
        return ctx

    @gen_test
    async def test_meta(self):
        response = await self.async_fetch("/unsafe/meta/800x400/image.jpg")
        expect(response.code).to_equal(200)

    @gen_test
    async def test_meta_with_unicode(self):
        response = await self.async_fetch(
            "/unsafe/meta/200x300/alabama1_ap620%C3%A9.jpg"
        )
        expect(response.code).to_equal(200)
        obj = loads(response.body.decode("utf-8"))
        expect(obj["thumbor"]["target"]["width"]).to_equal(200)
        expect(obj["thumbor"]["target"]["height"]).to_equal(300)

    @gen_test
    async def test_meta_frame_count(self):
        response = await self.async_fetch("/unsafe/meta/800x400/image.jpg")
        expect(response.code).to_equal(200)
        obj = loads(response.body.decode("utf-8"))
        expect(obj["thumbor"]["source"]["frameCount"]).to_equal(1)

    @gen_test
    async def test_meta_frame_count_with_gif(self):
        response = await self.async_fetch("/unsafe/meta/animated.gif")
        expect(response.code).to_equal(200)
        obj = loads(response.body.decode("utf-8"))
        expect(obj["thumbor"]["source"]["frameCount"]).to_equal(2)

    @gen_test
    async def test_max_bytes(self):
        response = await self.async_fetch(
            "/unsafe/filters:max_bytes(35000)/Giunchedi%2C_"
            "Filippo_January_2015_01.jpg"
        )
        expect(response.code).to_equal(200)
        expect(len(response.body)).to_be_lesser_or_equal_to(35000)

    @gen_test
    async def test_max_bytes_impossible(self):
        response = await self.async_fetch(
            "/unsafe/filters:max_bytes(1000)/Giunchedi%2C_Filippo_"
            "January_2015_01.jpg"
        )
        expect(response.code).to_equal(200)
        expect(len(response.body)).to_be_greater_than(1000)

    @gen_test
    async def test_meta_with_exif_orientation(self):
        response = await self.async_fetch(
            "/unsafe/meta/0x0/Giunchedi%2C_Filippo_January_2015_01-"
            "cmyk-orientation-exif.jpg"
        )
        expect(response.code).to_equal(200)
        obj = loads(response.body.decode("utf-8"))
        expect(obj["thumbor"]["target"]["width"]).to_equal(533)
        expect(obj["thumbor"]["target"]["height"]).to_equal(800)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import datetime
import functools
import re
import sys
import traceback

import pytz
import tornado.web
from tornado.locks import Condition

import thumbor.filters
from thumbor import __version__
from thumbor.context import Context, RequestParameters
from thumbor.engines import BaseEngine, EngineResult
from thumbor.engines.json_engine import JSONEngine
from thumbor.loaders import LoaderResult
from thumbor.result_storages import ResultStorageResult
from thumbor.storages.mixed_storage import Storage as MixedStorage
from thumbor.storages.no_storage import Storage as NoStorage
from thumbor.transformer import Transformer
from thumbor.utils import CONTENT_TYPE, EXTENSION, logger

HTTP_DATE_FMT = "%a, %d %b %Y %H:%M:%S GMT"

# Handlers should not override __init__ pylint: disable=attribute-defined-outside-init,arguments-differ
# pylint: disable=broad-except,abstract-method,too-many-branches,too-many-return-statements,too-many-statements


class FetchResult:  # Data Object pylint: disable=too-few-public-methods
    def __init__(
        self,
        normalized=False,
        buffer=None,
        engine=None,
        successful=False,
        loader_error=None,
    ):
        self.normalized = normalized
        self.engine = engine
        self.buffer = buffer
        self.successful = successful
        self.loader_error = loader_error
        self.exception = None


class BaseHandler(tornado.web.RequestHandler):
    url_locks = {}

    def prepare(self):
        super().prepare()

        if not hasattr(self, "context"):
            return

        self._response_ext = None
        self._response_length = None

        self._response_start = datetime.datetime.now()
        self.context.metrics.incr("response.count")

    def on_finish(self):
        super().on_finish()
        self.context = getattr(self, "context", None)

        if self.context is None:
            return

        if not hasattr(self.context, "request"):
            self.context.request = RequestParameters()

        total_time = (
            datetime.datetime.now() - self._response_start
        ).total_seconds() * 1000
        status = self.get_status()
        self.context.metrics.timing("response.time", total_time)
        self.context.metrics.timing(f"response.time.{status}", total_time)
        self.context.metrics.incr(f"response.status.{status}")

        if status == 200 and self.context is not None:
            if self.context.request.smart:
                self.context.metrics.incr("response.smart.count")
                self.context.metrics.timing(
                    "response.smart.latency", total_time
                )
            else:
                self.context.metrics.incr("response.not_smart.count")
                self.context.metrics.timing(
                    "response.not_smart.latency", total_time
                )

        if self._response_ext is not None:
            ext = self._response_ext
            self.context.metrics.incr(f"response.format{ext}")
            self.context.metrics.timing(f"response.time{ext}", total_time)

            if self._response_length is not None:
                self.context.metrics.incr(
                    f"response.bytes{ext}", self._response_length
                )

    def _error(self, status, msg=None):
        self.set_status(status)

        if msg is not None:
            logger.warning(msg)
        self.finish()

    async def execute_image_operations(self):
        self.context.request.quality = None

        req = self.context.request
        conf = self.context.config

        total_time = (
            datetime.datetime.now() - self._response_start
        ).total_seconds() * 1000

        if self.context.request.smart:
            self.context.metrics.incr("response.smart")
            self.context.metrics.timing("response.smart", total_time)
        else:
            self.context.metrics.incr("response.none_smart")
            self.context.metrics.timing("response.none_smart", total_time)

        should_store = (
            self.context.config.RESULT_STORAGE_STORES_UNSAFE
            or not self.context.request.unsafe
        )

        if self.context.modules.result_storage and should_store:
            start = datetime.datetime.now()

            try:
                result = await self.context.modules.result_storage.get()
            except Exception as error:
                logger.exception(
                    "[BaseHander.execute_image_operations] %s", error
                )
                self._error(
                    500,
                    "Error while trying to get the image "
                    f"from the result storage: {error}",
                )

                return

            finish = datetime.datetime.now()

            self.context.metrics.timing(
                "result_storage.incoming_time",
                (finish - start).total_seconds() * 1000,
            )

            if result is None:
                self.context.metrics.incr("result_storage.miss")
            else:
                self.context.metrics.incr("result_storage.hit")
                self.context.metrics.incr(
                    "result_storage.bytes_read", len(result)
                )
                logger.debug("[RESULT_STORAGE] IMAGE FOUND: %s", req.url)
                await self.finish_request(result)

                return

        if (
            conf.MAX_WIDTH
            and (not isinstance(req.width, str))
            and req.width > conf.MAX_WIDTH
        ):
            req.width = conf.MAX_WIDTH

        if (
            conf.MAX_HEIGHT
            and (not isinstance(req.height, str))
            and req.height > conf.MAX_HEIGHT
        ):
            req.height = conf.MAX_HEIGHT

        req.meta_callback = (
            conf.META_CALLBACK_NAME
            or self.request.arguments.get("callback", [None])[0]
        )

        self.filters_runner = self.context.filters_factory.create_instances(
            self.context, self.context.request.filters
        )
        # Apply all the filters from the PRE_LOAD phase
        # and call get_image() afterwards.
        await self.filters_runner.apply_filters(
            thumbor.filters.PHASE_PRE_LOAD,
        )
        await self.get_image()

    # TODO: refactor this
    async def get_image(self):
        """
        This function is called after the PRE_LOAD filters have been applied.
        It applies the AFTER_LOAD filters on the result, then crops the image.
        """
        try:
            result = await self._fetch(self.context.request.image_url)

            if not result.successful:
                if result.loader_error == LoaderResult.ERROR_NOT_FOUND:
                    self._error(404)

                    return

                if result.loader_error == LoaderResult.ERROR_UPSTREAM:
                    # Return a Bad Gateway status if the error
                    # came from upstream
                    self._error(502)

                    return

                if result.loader_error == LoaderResult.ERROR_TIMEOUT:
                    # Return a Gateway Timeout status if upstream
                    # timed out (i.e. 599)
                    self._error(504)

                    return

                if isinstance(result.loader_error, int):
                    self._error(result.loader_error)

                    return

                if (
                    hasattr(result, "engine_error")
                    and result.engine_error
                    == EngineResult.COULD_NOT_LOAD_IMAGE
                ):
                    self._error(400)

                    return

                self._error(500)

                return

        except Exception as error:
            req_url = self.context.request.image_url
            msg = f"[BaseHandler] get_image failed for url `{req_url}`. error: `{error}`"

            self.log_exception(*sys.exc_info())

            if "cannot identify image file" in str(error):
                logger.warning(msg)
                self._error(400)
            else:
                logger.error(msg)
                self._error(500)

            return

        normalized = result.normalized
        buffer = result.buffer
        engine = result.engine

        req = self.context.request

        if engine is None:
            if buffer is None:
                self._error(504)

                return

            engine = self.context.request.engine
            try:
                engine.load(buffer, self.context.request.extension)
            except Exception as error:
                logger.exception("Loading image failed with %s", error)
                self._error(504)

                return

        self.context.transformer = Transformer(self.context)

        await self.filters_runner.apply_filters(
            thumbor.filters.PHASE_AFTER_LOAD
        )
        self.normalize_crops(normalized, req, engine)

        if req.meta:
            self.context.transformer.engine = (
                self.context.request.engine
            ) = JSONEngine(engine, req.image_url, req.meta_callback)

        await self.context.transformer.transform()
        await self.after_transform()

    def normalize_crops(self, normalized, req, engine):
        new_crops = None

        if normalized and req.should_crop:
            crop_left = req.crop["left"]
            crop_top = req.crop["top"]
            crop_right = req.crop["right"]
            crop_bottom = req.crop["bottom"]

            actual_width, actual_height = engine.size

            if not req.width and not req.height:
                actual_width = engine.size[0]
                actual_height = engine.size[1]
            elif req.width:
                actual_height = engine.get_proportional_height(engine.size[0])
            elif req.height:
                actual_width = engine.get_proportional_width(engine.size[1])

            new_crops = self.translate_crop_coordinates(
                engine.source_width,
                engine.source_height,
                actual_width,
                actual_height,
                crop_left,
                crop_top,
                crop_right,
                crop_bottom,
            )
            req.crop["left"] = new_crops[0]
            req.crop["top"] = new_crops[1]
            req.crop["right"] = new_crops[2]
            req.crop["bottom"] = new_crops[3]

    async def after_transform(self):
        if (
            self.context.request.extension != ".gif"
            or self.context.config.USE_GIFSICLE_ENGINE is None
        ):
            await self.filters_runner.apply_filters(
                thumbor.filters.PHASE_POST_TRANSFORM
            )

        await self.finish_request()

    def is_webp(self, context):
        return (
            context.config.AUTO_WEBP
            and context.request.accepts_webp
            and not context.request.engine.is_multiple()
            and context.request.engine.can_convert_to_webp()
        )

    def is_animated_gif(self, data):
        if data[:6] not in [b"GIF87a", b"GIF89a"]:
            return False
        i = 10  # skip header
        frames = 0

        def skip_color_table(i, flags):
            if flags & 0x80:
                i += 3 << ((flags & 7) + 1)

            return i

        flags = data[i]
        i = skip_color_table(i + 3, flags)

        while frames < 2:
            block = data[i : i + 1]  # NOQA
            i += 1

            if block == b"\x3B":
                break

            if block == b"\x21":
                i += 1
            elif block == b"\x2C":
                frames += 1
                i += 8
                i = skip_color_table(i + 1, data[i])
                i += 1
            else:
                return False

            while True:
                j = data[i]
                i += 1

                if not j:
                    break
                i += j

        return frames > 1

    def can_auto_convert_png_to_jpg(self):
        request_override = self.context.request.auto_png_to_jpg
        config = self.context.config

        enabled = (
            request_override is None
            if config.AUTO_PNG_TO_JPG
            else request_override
        )

        if enabled:
            return self.context.request.engine.can_auto_convert_png_to_jpg()

        return False

    def define_image_type(self, context, result):
        if result is not None:
            if isinstance(result, ResultStorageResult):
                buffer = result.buffer
            else:
                buffer = result
            image_extension = EXTENSION.get(
                BaseEngine.get_mimetype(buffer), ".jpg"
            )
            content_type = CONTENT_TYPE.get(
                image_extension, CONTENT_TYPE[".jpg"]
            )

            return image_extension, content_type

        image_extension = context.request.format

        if image_extension is not None:
            image_extension = f".{image_extension}"
            logger.debug("Image format specified as %s.", image_extension)
        elif self.is_webp(context):
            image_extension = ".webp"
            logger.debug(
                "Image format set by AUTO_WEBP as %s.", image_extension
            )
        elif self.can_auto_convert_png_to_jpg():
            image_extension = ".jpg"
            logger.debug(
                "Image format set by AUTO_PNG_TO_JPG as %s.",
                image_extension,
            )
        else:
            image_extension = context.request.engine.extension
            logger.debug(
                "No image format specified. Retrieving "
                "from the image extension: %s.",
                image_extension,
            )

        content_type = CONTENT_TYPE.get(image_extension, CONTENT_TYPE[".jpg"])

        if context.request.meta:
            context.request.meta_callback = (
                context.config.META_CALLBACK_NAME
                or self.request.arguments.get("callback", [None])[0]
            )
            content_type = (
                "text/javascript"
                if context.request.meta_callback
                else "application/json"
            )
            logger.debug(
                "Metadata requested. Serving content type of %s.", content_type
            )

        logger.debug("Content Type of %s detected.", content_type)

        return (image_extension, content_type)

    def _load_results(self, context):
        image_extension, content_type = self.define_image_type(context, None)

        quality = self.context.request.quality

        if quality is None:
            if (
                image_extension == ".webp"
                and self.context.config.WEBP_QUALITY is not None
            ):
                quality = self.context.config.get("WEBP_QUALITY")
            elif (
                image_extension == ".avif"
                and self.context.config.AVIF_QUALITY is not None
            ):
                quality = self.context.config.AVIF_QUALITY
            elif image_extension in (".heif", ".heic"):
                quality = self.context.config.HEIF_QUALITY
                if quality is None:
                    quality = 75
            else:
                quality = self.context.config.QUALITY

        results = context.request.engine.read(image_extension, quality)

        if context.request.max_bytes is not None:
            results = self.reload_to_fit_in_kb(
                context.request.engine,
                results,
                image_extension,
                quality,
                context.request.max_bytes,
            )

        if not context.request.meta:
            results = self.optimize(context, image_extension, results)
            # An optimizer might have modified the image format.
            content_type = BaseEngine.get_mimetype(results)

        return results, content_type

    async def _process_result_from_storage(self, result):
        if self.context.config.SEND_IF_MODIFIED_LAST_MODIFIED_HEADERS:
            # Handle If-Modified-Since & Last-Modified header
            try:
                if isinstance(result, ResultStorageResult):
                    result_last_modified = result.last_modified
                else:
                    result_last_modified = await (
                        self.context.modules.result_storage.last_updated()
                    )

                if result_last_modified:
                    if "If-Modified-Since" in self.request.headers:
                        date_modified_since = datetime.datetime.strptime(
                            self.request.headers["If-Modified-Since"],
                            HTTP_DATE_FMT,
                        ).replace(tzinfo=pytz.utc)

                        if result_last_modified <= date_modified_since:
                            self.set_status(304)
                            self.finish()

                            return

                    self.set_header(
                        "Last-Modified",
                        result_last_modified.strftime(HTTP_DATE_FMT),
                    )
            except NotImplementedError:
                logger.warning(
                    "last_updated method is not supported by your "
                    "result storage service, hence If-Modified-Since & "
                    "Last-Updated headers support is disabled."
                )

    async def finish_request(self, result_from_storage=None):
        if result_from_storage is not None:
            await self._process_result_from_storage(result_from_storage)

            _, content_type = self.define_image_type(
                self.context, result_from_storage
            )
            await self._write_results_to_client(
                result_from_storage, content_type
            )

            return

        context = self.context
        result_storage = context.modules.result_storage
        metrics = context.metrics

        should_store = (
            result_storage
            and not context.request.prevent_result_storage
            and (
                context.config.RESULT_STORAGE_STORES_UNSAFE
                or not context.request.unsafe
            )
        )

        try:
            result = await self.context.thread_pool.queue(
                operation=functools.partial(self._load_results, context),
            )
        except Exception as error:  # pylint: disable=broad-except
            logger.exception("[BaseHander.finish_request] %s", error)
            self._error(500, f"Error while trying to fetch the image: {error}")

            return

        (results, content_type) = result
        await self._write_results_to_client(results, content_type)

        if should_store:
            await self._store_results(result_storage, metrics, results)

        # can't cleanup before storing results as the storage requires context
        self._cleanup()

    def _cleanup(self):
        self.context.request_handler = None

        if hasattr(self.context, "request"):
            self.context.request.engine = None
        self.context.modules = None
        self.context.filters_factory = None
        self.context.metrics = None
        self.context.thread_pool = None
        self.context.transformer = None
        self.context = None  # Handlers should not override __init__ pylint: disable=attribute-defined-outside-init

    async def _write_results_to_client(self, results, content_type):
        max_age = self.context.config.MAX_AGE

        if self.context.request.max_age is not None:
            max_age = self.context.request.max_age

        if (
            self.context.request.prevent_result_storage
            or self.context.request.detection_error
        ):
            max_age = self.context.config.MAX_AGE_TEMP_IMAGE

        if max_age:
            self.set_header(
                "Cache-Control", "max-age=" + str(max_age) + ",public"
            )
            self.set_header(
                "Expires",
                datetime.datetime.utcnow()
                + datetime.timedelta(seconds=max_age),
            )

        if hasattr(self.context.config, "ACCESS_CONTROL_ALLOW_ORIGIN_HEADER"):
            ac_header = self.context.config.ACCESS_CONTROL_ALLOW_ORIGIN_HEADER

            if ac_header is not False:
                self.set_header("Access-Control-Allow-Origin", ac_header)
                logger.debug("CORS header found. Set to: %s", ac_header)

        self.set_header("Server", f"Thumbor/{__version__}")
        self.set_header("Content-Type", content_type)

        if isinstance(results, ResultStorageResult):
            buffer = results.buffer
        else:
            buffer = results

        # auto-convert configured?
        should_vary = self.context.config.AUTO_WEBP
        # we have image (not video)
        should_vary = should_vary and content_type.startswith("image/")
        # output format is not requested via format filter
        should_vary = should_vary and not (
            self.context.request.format
            and bool(  # format is supported by filter
                re.search(r"format\([^)]+\)", self.context.request.filters)
            )  # filter is in request
        )
        # our image is not animated gif
        should_vary = should_vary and not self.is_animated_gif(buffer)

        if should_vary:
            self.set_header("Vary", "Accept")

        self.context.headers = self._headers.copy()
        self._response_ext = EXTENSION.get(content_type)
        self._response_length = len(buffer)

        self.write(buffer)
        self.finish()

    async def _store_results(self, result_storage, metrics, results):
        start = datetime.datetime.now()

        await result_storage.put(results)

        finish = datetime.datetime.now()
        metrics.incr("result_storage.bytes_written", len(results))
        metrics.timing(
            "result_storage.outgoing_time",
            (finish - start).total_seconds() * 1000,
        )

    def optimize(self, context, image_extension, results):
        for optimizer in context.modules.optimizers:
            new_results = optimizer(context).run_optimizer(
                image_extension, results
            )

            if new_results is not None:
                results = new_results

        return results

    @staticmethod
    def reload_to_fit_in_kb(
        engine, initial_results, extension, initial_quality, max_bytes
    ):
        if (
            extension not in [".webp", ".jpg", ".jpeg"]
            or len(initial_results) <= max_bytes
        ):
            return initial_results

        results = initial_results
        quality = initial_quality

        while len(results) > max_bytes:
            quality = int(quality * 0.75)

            if quality < 10:
                logger.debug(
                    "Could not find any reduction that matches "
                    "required size of %d bytes.",
                    max_bytes,
                )

                return initial_results

            logger.debug(
                "Trying to downsize image with quality of %d...", quality
            )
            results = engine.read(extension, quality)

        prev_result = results

        while len(results) <= max_bytes and quality < initial_quality:
            quality = max(initial_quality, int(quality * 1.1))
            logger.debug(
                "Trying to upsize image with quality of %d...", quality
            )
            prev_result = results
            results = engine.read(extension, quality)

        return prev_result

    @classmethod
    def translate_crop_coordinates(
        cls,
        original_width,
        original_height,
        width,
        height,
        crop_left,
        crop_top,
        crop_right,
        crop_bottom,
    ):

        if original_width == width and original_height == height:
            return None

        crop_left = crop_left * width // original_width
        crop_top = crop_top * height // original_height

        crop_right = crop_right * width // original_width
        crop_bottom = crop_bottom * height // original_height

        return (crop_left, crop_top, crop_right, crop_bottom)

    def validate(self, path):
        if not hasattr(self.context.modules.loader, "validate"):
            return True

        is_valid = self.context.modules.loader.validate(self.context, path)

        if not is_valid:
            logger.warning(
                (
                    'Request denied because the specified path "%s" was '
                    "not identified by the loader as a valid path"
                ),
                path,
            )

        return is_valid

    async def _fetch(self, url):
        """

        :param url:
        :type url:
        :return:
        :rtype:
        """
        fetch_result = FetchResult()

        storage = self.context.modules.storage

        await self.acquire_url_lock(url)

        try:
            fetch_result.buffer = await storage.get(url)
            mime = None

            if fetch_result.buffer is not None:
                self.release_url_lock(url)

                fetch_result.successful = True

                self.context.metrics.incr("storage.hit")

                mime = BaseEngine.get_mimetype(fetch_result.buffer)
                self.context.request.extension = EXTENSION.get(mime, ".jpg")

                if (
                    mime == "image/gif"
                    and self.context.config.USE_GIFSICLE_ENGINE
                ):
                    self.context.request.engine = (
                        self.context.modules.gif_engine
                    )
                else:
                    self.context.request.engine = self.context.modules.engine

                return fetch_result

            self.context.metrics.incr("storage.miss")

            loader_result = await self.context.modules.loader.load(
                self.context, url
            )
        finally:
            self.release_url_lock(url)

        if isinstance(loader_result, LoaderResult):
            # TODO _fetch should probably return a result object vs a list to
            # to allow returning metadata

            if not loader_result.successful:
                fetch_result.buffer = None
                fetch_result.loader_error = loader_result.error

                return fetch_result

            fetch_result.buffer = loader_result.buffer
        else:
            # Handle old loaders
            fetch_result.buffer = loader_result

        if fetch_result.buffer is None:
            return fetch_result

        fetch_result.successful = True

        if mime is None:
            mime = BaseEngine.get_mimetype(fetch_result.buffer)

        self.context.request.extension = extension = EXTENSION.get(
            mime, ".jpg"
        )

        try:
            if mime == "image/gif" and self.context.config.USE_GIFSICLE_ENGINE:
                self.context.request.engine = self.context.modules.gif_engine
            else:
                self.context.request.engine = self.context.modules.engine

            self.context.request.engine.load(fetch_result.buffer, extension)

            if self.context.request.engine.image is None:
                fetch_result.successful = False
                fetch_result.buffer = None
                fetch_result.engine = self.context.request.engine
                fetch_result.engine_error = EngineResult.COULD_NOT_LOAD_IMAGE

                return fetch_result

            fetch_result.normalized = self.context.request.engine.normalize()

            # Allows engine or loader to override storage
            # on the fly for the purpose of
            # marking a specific file as unstoreable
            storage = self.context.modules.storage

            is_no_storage = isinstance(storage, NoStorage)
            is_mixed_storage = isinstance(storage, MixedStorage)
            is_mixed_no_file_storage = is_mixed_storage and isinstance(
                storage.file_storage, NoStorage
            )

            if not (is_no_storage or is_mixed_no_file_storage):
                await storage.put(url, fetch_result.buffer)

            await storage.put_crypto(url)
        except Exception as error:
            fetch_result.successful = False
            fetch_result.exception = error

        if not fetch_result.successful and fetch_result.exception is not None:
            raise fetch_result.exception
        fetch_result.buffer = None
        fetch_result.engine = self.context.request.engine

        return fetch_result

    async def get_blacklist_contents(self):
        filename = "blacklist.txt"

        exists = await (self.context.modules.storage.exists(filename))

        if exists:
            blacklist = await (self.context.modules.storage.get(filename))

            return blacklist.decode()

        return ""

    async def acquire_url_lock(self, url):
        if url not in BaseHandler.url_locks:
            BaseHandler.url_locks[url] = Condition()
        else:
            return await BaseHandler.url_locks[url].wait()

    def release_url_lock(self, url):
        try:
            BaseHandler.url_locks[url].notify_all()
            del BaseHandler.url_locks[url]
        except KeyError:
            pass


class ContextHandler(BaseHandler):
    def initialize(self, context):
        self.context = Context(
            server=context.server,
            config=context.config,
            importer=context.modules.importer,
            request_handler=self,
        )
        self.context.metrics.initialize(self)

    def log_exception(self, *exc_info):
        if isinstance(exc_info[1], tornado.web.HTTPError):
            # Delegate HTTPError's to the base class
            # We don't want these through normal exception handling
            super().log_exception(*exc_info)

            return

        msg = traceback.format_exception(  # pylint: disable=no-value-for-parameter
            *exc_info
        )

        try:
            if self.context.config.USE_CUSTOM_ERROR_HANDLING:
                self.context.modules.importer.error_handler.handle_error(
                    context=self.context, handler=self, exception=exc_info
                )
        finally:
            del exc_info
            logger.error("ERROR: %s", "".join(msg))


##
# Base handler for Image API operations
##
class ImageApiHandler(ContextHandler):
    def validate(self, body):  # pylint: disable=arguments-renamed
        conf = self.context.config
        mime = BaseEngine.get_mimetype(body)

        if mime == "image/gif" and self.context.config.USE_GIFSICLE_ENGINE:
            engine = self.context.modules.gif_engine
        else:
            engine = self.context.modules.engine

        # Check if image is valid
        try:
            engine.load(body, None)
        except IOError:
            self._error(415, "Unsupported Media Type")

            return False

        # Check weight constraints

        if (
            conf.UPLOAD_MAX_SIZE != 0
            and len(self.request.body) > conf.UPLOAD_MAX_SIZE
        ):
            self._error(
                412,
                "Image exceed max weight "
                f"(Expected : {conf.UPLOAD_MAX_SIZE}, Actual : {len(self.request.body)})",
            )

            return False

        # Check size constraints
        size = engine.size

        if conf.MIN_WIDTH > size[0] or conf.MIN_HEIGHT > size[1]:
            self._error(
                412,
                "Image is too small (Expected: %s/%s , Actual : %s/%s) % "
                "(conf.MIN_WIDTH, conf.MIN_HEIGHT, size[0], size[1])",
            )

            return False

        return True

    async def write_file(self, file_id, body):
        storage = self.context.modules.upload_photo_storage
        await storage.put(file_id, body)



# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2022, globo.com <thumbor@g.globo>

from io import BytesIO

from PIL import Image
from preggy import expect
from tornado.testing import gen_test

from tests.handlers.test_base_handler import BaseImagingTestCase
from thumbor.config import Config
from thumbor.context import Context, ServerParameters
from thumbor.engines.pil import Engine
from thumbor.importer import Importer

# pylint: disable=broad-except,abstract-method,attribute-defined-outside-init,line-too-long,too-many-public-methods
# pylint: disable=too-many-lines


class ImageOperationsWithKeepExifCopyrightTags(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path
        cfg.STORAGE = "thumbor.storages.no_storage"
        cfg.RESPECT_ORIENTATION = True

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        self.context = Context(server, cfg, importer)
        return self.context

    @gen_test
    async def test_should_not_return_exif_copyright_tags_with_copyright_tags(
        self,
    ):
        self.context.config.PRESERVE_EXIF_COPYRIGHT_INFO = False
        self.context.config.PRESERVE_EXIF_INFO = False
        response = await self.async_fetch(
            "/unsafe/filters:strip_exif()/thumbor-exif.png"
        )
        expect(response.code).to_equal(200)
        ctx = self.context
        engine = Engine(ctx)
        engine.load(response.body, ".jpg")
        final_bytes = BytesIO(engine.read())
        image = Image.open(final_bytes)
        expect(image.info.get("exif")).to_be_null()

    @gen_test
    async def test_should_return_exif_copyright_tags(self):
        self.context.config.PRESERVE_EXIF_COPYRIGHT_INFO = True
        response = await self.async_fetch(
            "/unsafe/filters:strip_exif()/thumbor-exif.png"
        )
        expect(response.code).to_equal(200)

        engine = Engine(self.context)
        engine.load(response.body, ".jpg")
        final_bytes = BytesIO(engine.read())
        image = Image.open(final_bytes)
        expect(image.info.get("exif")).not_to_be_null()

    @gen_test
    async def test_should_not_return_exif_copyright_tags(self):
        self.context.config.PRESERVE_EXIF_COPYRIGHT_INFO = True
        response = await self.async_fetch("/unsafe/20x20.jpg")
        expect(response.code).to_equal(200)
        engine = Engine(self.context)
        engine.load(response.body, ".jpg")
        final_bytes = BytesIO(engine.read())
        image = Image.open(final_bytes)

        expect(image.info.get("exif")).to_be_null()



#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging
from itertools import product

from colorama import Fore

DEBUGS = ["", "debug"]

METAS = ["meta"]

TRIMS = [
    "trim",
    "trim:top-left",
    "trim:bottom-right",
    "trim:top-left:10",
    "trim:bottom-right:20",
]

CROPS = ["10x10:100x100"]

FITINS = ["fit-in", "adaptive-fit-in", "full-fit-in", "adaptive-full-fit-in"]

SIZES = [
    "200x200",
    "-300x100",
    "100x-300",
    "-100x-300",
    "origx300",
    "200xorig",
    "origxorig",
]

H_ALIGNS = [
    "left",
    "right",
    "center",
]

V_ALIGNS = [
    "top",
    "bottom",
    "middle",
]

SMARTS = [
    "smart",
]

FILTERS = [
    "filters:brightness(10)",
    "filters:contrast(10)",
    "filters:equalize()",
    "filters:grayscale()",
    "filters:rotate(90)",
    "filters:noise(10)",
    "filters:quality(5)",
    "filters:redeye()",
    "filters:rgb(10,-10,20)",
    "filters:round_corner(20,255,255,100)",
    "filters:sharpen(6,2.5,false)",
    "filters:sharpen(6,2.5,true)",
    "filters:strip_exif()",
    "filters:strip_icc()",
    "filters:watermark(rgba-interlaced.png,10,10,50)",
    "filters:watermark(rgba-interlaced.png,center,center,50)",
    "filters:watermark(rgba-interlaced.png,repeat,repeat,50)",
    "filters:frame(rgba.png)",
    "filters:fill(ff0000)",
    "filters:fill(auto)",
    "filters:fill(ff0000,true)",
    "filters:fill(transparent)",
    "filters:fill(transparent,true)",
    "filters:blur(2)",
    "filters:extract_focal()",
    "filters:focal()",
    "filters:focal(0x0:1x1)",
    "filters:no_upscale()",
    "filters:gifv()",
    "filters:gifv(webm)",
    "filters:gifv(mp4)",
    "filters:max_age(600)",
    "filters:upscale()",
    "filters:format(avif)",
    "filters:format(heic)",
    "filters:format(heif)",
    # one big filter 4-line string
    "filters:curve([(0,0),(255,255)],[(0,50),(16,51),(32,69),"
    "(58,85),(92,120),(128,170),(140,186),(167,225),"  # NOQA
    "(192,245),(225,255),(244,255),(255,254)],[(0,0),(16,2),"
    "(32,18),(64,59),(92,116),(128,182),(167,211),(192,227)"  # NOQA
    ",(224,240),(244,247),(255,252)],[(0,48),(16,50),(62,77),"
    "(92,110),(128,144),(140,153),(167,180),(192,192),"  # NOQA
    "(224,217),(244,225),(255,225)])",
]

ORIGINAL_IMAGES_BASE = [
    "gradient.jpg",
    "cmyk.jpg",
    "rgba.png",
    "grayscale.jpg",
    "16bit.png",
    "thumbor-exif.png",
]

ORIGINAL_IMAGES_GIF_WEBP = [
    "gradient.webp",
    "gradient.gif",
    "animated.gif",
]

ALL_OPTIONS = (
    METAS
    + TRIMS
    + CROPS
    + FITINS
    + SIZES
    + H_ALIGNS
    + V_ALIGNS
    + SMARTS
    + FILTERS
)

MAX_DATASET_SIZE = len(ALL_OPTIONS) * (
    len(ORIGINAL_IMAGES_BASE) + len(ORIGINAL_IMAGES_GIF_WEBP)
)


class UrlsTester:
    def __init__(self, http_client):
        self.failed_items = []
        self.http_client = http_client

    def report(self):
        if len(self.failed_items) == 0:
            return

        raise AssertionError("Failed urls:\n%s" % "\n".join(self.failed_items))

    async def try_url(self, url):
        result = None
        error = None
        failed = False

        try:
            result = await self.http_client.fetch(url, request_timeout=60)
        except Exception as err:  # pylint: disable=broad-except
            logging.exception("Error in %s: %s", url, err)
            error = err
            failed = True

        if result is not None and result.code == 200 and not failed:
            print("{0.GREEN} SUCCESS ({1}){0.RESET}".format(Fore, url))
            return

        self.failed_items.append(url)
        print(
            "{0.RED} FAILED ({1}) - ERR({2}) {0.RESET}".format(
                Fore, url, result is not None and result.code or error
            )
        )


def single_dataset(with_gif=True):
    images = ORIGINAL_IMAGES_BASE[:]
    if with_gif:
        images += ORIGINAL_IMAGES_GIF_WEBP
    return product(ALL_OPTIONS, images)


def combined_dataset(with_gif=True):
    images = ORIGINAL_IMAGES_BASE[:]
    if with_gif:
        images += ORIGINAL_IMAGES_GIF_WEBP
    combined_options = product(
        TRIMS[:2],
        CROPS[:2],
        FITINS[:2],
        SIZES[:2],
        H_ALIGNS[:2],
        V_ALIGNS[:2],
        SMARTS[:2],
        FILTERS[:2],
        images,
    )
    return combined_options



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from json import loads
from unittest import mock

from preggy import expect
from redis import Redis, Sentinel
from tornado.testing import gen_test

from tests.base import DetectorTestCase
from thumbor.config import Config
from thumbor.detectors.queued_detector import QueuedDetector

TEST_REDIS_HOST = "0.0.0.0"
TEST_REDIS_PORT = 6668
TEST_REDIS_PASSWORD = "hey_you"
TEST_REDIS_MODE = "single_node"
TEST_REDIS_SENTINEL_MODE = "sentinel"
TEST_REDIS_QUEUE_SENTINEL_INSTANCES = "localhost:26379"
TEST_REDIS_SENTINEL_MASTER_INSTANCE = "masterinstance"
TEST_REDIS_SENTINEL_SOCKET_TIMEOUT = 10.0


class SharedQueuedDetectorTestCase(DetectorTestCase):
    def setUp(self):
        super().setUp()
        self.redis = None

    async def detector_send_to_queues(self, ctx):
        detector = QueuedDetector(ctx, 1, [])
        expect(detector).not_to_be_null()

        data = await detector.detect()
        expect(data).to_be_empty()
        expect(ctx.request.detection_error).to_be_false()

        result = self.redis.get("resque:unique:queue:Detect:/image/test.jpg")
        expect(result).to_equal("1")

        expected_payload = {
            "queue": "Detect",
            "args": ["all", "/image/test.jpg", "/image/test.jpg"],
            "class": "remotecv.pyres_tasks.DetectTask",
            "key": "/image/test.jpg",
        }

        result = self.redis.lpop("resque:queue:Detect")
        expect(loads(result.decode("utf-8"))).to_be_like(expected_payload)

    async def detector_fails_properly(self, ctx):
        detector = QueuedDetector(ctx, 1, [])
        expect(detector).not_to_be_null()

        data = await detector.detect()
        expect(data).to_be_empty()
        expect(ctx.request.detection_error).to_be_true()
        expect(detector.queue).to_be_null()

    async def detector_can_detect_twice(self, ctx):
        detector = QueuedDetector(ctx, 1, [])
        expect(detector).not_to_be_null()

        data = await detector.detect()
        expect(data).to_be_empty()
        expect(ctx.request.detection_error).to_be_false()
        expect(detector.queue).not_to_be_null()

        data = detector.detect()
        expect(detector.queue).not_to_be_null()


class QueuedDetectorTestCase(SharedQueuedDetectorTestCase):
    def get_config(self):
        return Config(
            REDIS_QUEUE_SERVER_PORT=TEST_REDIS_PORT,
            REDIS_QUEUE_SERVER_PASSWORD=TEST_REDIS_PASSWORD,
        )

    def setUp(self):
        super().setUp()
        self.redis = Redis(
            host=TEST_REDIS_HOST,
            port=TEST_REDIS_PORT,
            db=0,
            password=TEST_REDIS_PASSWORD,
        )

        self.redis.delete("resque:unique:queue:Detect:/image/test.jpg")
        self.redis.delete("resque:queue:Detect")
        QueuedDetector.queue = None

    @gen_test
    async def test_detector_sends_to_queue(self):
        ctx = mock.Mock(
            config=mock.Mock(
                REDIS_QUEUE_MODE=TEST_REDIS_MODE,
                REDIS_QUEUE_SERVER_HOST=TEST_REDIS_HOST,
                REDIS_QUEUE_SERVER_PORT=TEST_REDIS_PORT,
                REDIS_QUEUE_SERVER_DB=0,
                REDIS_QUEUE_SERVER_PASSWORD=TEST_REDIS_PASSWORD,
            ),
            request=mock.Mock(
                image_url="/image/test.jpg",
                detection_error=False,
            ),
        )
        await self.detector_send_to_queues(ctx)

    @gen_test
    async def test_detector_fails_properly(self):
        ctx = mock.Mock(
            config=mock.Mock(
                REDIS_QUEUE_MODE=TEST_REDIS_MODE,
                REDIS_QUEUE_SERVER_HOST=TEST_REDIS_HOST,
                REDIS_QUEUE_SERVER_PORT=6669,
                REDIS_QUEUE_SERVER_DB=0,
                REDIS_QUEUE_SERVER_PASSWORD=TEST_REDIS_PASSWORD,
            ),
            request=mock.Mock(
                image_url="/image/test.jpg",
                detection_error=False,
            ),
        )
        await self.detector_fails_properly(ctx)

    @gen_test
    async def test_detector_can_detect_twice(self):
        ctx = mock.Mock(
            config=mock.Mock(
                REDIS_QUEUE_MODE=TEST_REDIS_MODE,
                REDIS_QUEUE_SERVER_HOST=TEST_REDIS_HOST,
                REDIS_QUEUE_SERVER_PORT=TEST_REDIS_PORT,
                REDIS_QUEUE_SERVER_DB=0,
                REDIS_QUEUE_SERVER_PASSWORD=TEST_REDIS_PASSWORD,
            ),
            request=mock.Mock(
                image_url="/image/test.jpg",
                detection_error=False,
            ),
        )
        await self.detector_can_detect_twice(ctx)


class QueuedSentinelDetectorTestCase(SharedQueuedDetectorTestCase):
    def get_config(self):
        return Config(
            REDIS_QUEUE_MODE=TEST_REDIS_SENTINEL_MODE,
            REDIS_QUEUE_SENTINEL_PASSWORD=TEST_REDIS_PASSWORD,
            REDIS_QUEUE_SENTINEL_SOCKET_TIMEOUT=TEST_REDIS_SENTINEL_SOCKET_TIMEOUT,
            REDIS_QUEUE_SENTINEL_MASTER_INSTANCE=TEST_REDIS_SENTINEL_MASTER_INSTANCE,
            REDIS_QUEUE_SENTINEL_MASTER_PASSWORD=TEST_REDIS_PASSWORD,
            REDIS_QUEUE_SENTINEL_MASTER_DB=0,
        )

    def setUp(self):
        super().setUp()
        self.sentinel = Sentinel(
            [("localhost", 26379)],
            sentinel_kwargs={"password": TEST_REDIS_PASSWORD},
        )
        self.redis = self.sentinel.master_for(
            TEST_REDIS_SENTINEL_MASTER_INSTANCE,
            password=TEST_REDIS_PASSWORD,
            db=0,
        )

        self.redis.delete("resque:unique:queue:Detect:/image/test.jpg")
        self.redis.delete("resque:queue:Detect")
        QueuedDetector.queue = None

        self.request = mock.Mock(
            image_url="/image/test.jpg",
            detection_error=False,
        )

        self.ctx = mock.Mock(
            config=self.config,
            request=self.request,
        )

    @gen_test
    async def test_detector_sends_to_queue(self):
        await self.detector_send_to_queues(self.ctx)

    @gen_test
    async def test_detector_fails_properly(self):
        self.ctx.config.REDIS_QUEUE_SENTINEL_INSTANCES = "localhost:23680"
        await self.detector_fails_properly(self.ctx)

    @gen_test
    async def test_detector_can_detect_twice(self):
        await self.detector_can_detect_twice(self.ctx)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com


from preggy import expect
from tornado.testing import gen_test

from tests.base import TestCase
from thumbor.handlers import BaseHandler

# pylint: disable=broad-except,abstract-method,attribute-defined-outside-init,line-too-long,too-many-public-methods
# pylint: disable=too-many-lines


class TranslateCoordinatesTestCase(TestCase):
    def setUp(self):
        super().setUp()
        coords = self.get_coords()
        crop_coords = BaseHandler.translate_crop_coordinates(
            original_width=coords["original_width"],
            original_height=coords["original_height"],
            width=coords["width"],
            height=coords["height"],
            crop_left=coords["crop_left"],
            crop_top=coords["crop_top"],
            crop_right=coords["crop_right"],
            crop_bottom=coords["crop_bottom"],
        )
        self.translate_crop_coordinates = crop_coords

    def get_coords(self):
        return {
            "original_width": 3000,
            "original_height": 2000,
            "width": 1200,
            "height": 800,
            "crop_left": 100,
            "crop_top": 100,
            "crop_right": 200,
            "crop_bottom": 200,
            "expected_crop": (40, 40, 80, 80),
        }

    @gen_test
    async def test_should_be_a_list_of_coords(self):
        expect(self.translate_crop_coordinates).to_be_instance_of(tuple)

    @gen_test
    async def test_should_translate_from_original_to_resized(self):
        expect(self.translate_crop_coordinates).to_equal(
            self.get_coords()["expected_crop"]
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com


import os
from tempfile import NamedTemporaryFile


class BaseOptimizer:
    def __init__(self, context):
        self.context = context

    def should_run(
        self, image_extension, image_buffer
    ):  # pylint: disable=unused-argument
        return True

    def run_optimizer(self, image_extension, buffer):
        if not self.should_run(image_extension, buffer):
            return buffer

        ifile = NamedTemporaryFile(  # pylint: disable=consider-using-with
            delete=False
        )
        ofile = NamedTemporaryFile(  # pylint: disable=consider-using-with
            delete=False
        )
        try:
            ifile.write(buffer)
            ifile.close()
            ofile.close()

            self.optimize(buffer, ifile.name, ofile.name)

            with open(
                ofile.name, "rb"
            ) as output_file:  # reopen with file thats been changed with the optimizer
                return output_file.read()
        finally:
            os.unlink(ifile.name)
            os.unlink(ofile.name)

    def optimize(self, image_buffer, input_file, output_file):
        raise NotImplementedError()



import signal
import sys
import time

import tornado.ioloop

from thumbor.utils import logger


def signal_handler(server, config, sig, _):
    io_loop = tornado.ioloop.IOLoop.instance()

    def stop_loop(now, deadline):
        if now < deadline and (
            io_loop._callbacks  # pylint: disable=protected-access
            or io_loop._timeouts  # pylint: disable=protected-access
        ):
            logger.debug("Waiting for next tick")
            now += 1
            io_loop.add_timeout(now, stop_loop, now, deadline)
        else:
            io_loop.stop()
            logger.debug("Shutdown finally")

    def stop_server(now, deadline):
        if now < deadline:
            logger.debug("Waiting for next tick")
            now += 1
            io_loop.add_timeout(now, stop_server, now, deadline)
        else:
            server.stop()
            logger.debug("Stopped http server.")
            logger.debug(
                "Will shutdown io in maximum %d seconds ...",
                config.MAX_WAIT_SECONDS_BEFORE_IO_SHUTDOWN,
            )
            now = time.time()
            stop_loop(now, now + config.MAX_WAIT_SECONDS_BEFORE_IO_SHUTDOWN)

    def shutdown():
        logger.debug(
            "Stopping http server (i.e. stopping accepting new requests) in %d seconds ...",
            config.MAX_WAIT_SECONDS_BEFORE_SERVER_SHUTDOWN,
        )
        now = time.time()
        stop_server(now, now + config.MAX_WAIT_SECONDS_BEFORE_SERVER_SHUTDOWN)

    logger.warning("Caught signal: %d. Shutting down server.", sig)
    io_loop.add_callback_from_signal(shutdown)


def setup_signal_handler(server, config):
    def server_sig_handler(sig, frame):
        signal_handler(server, config, sig, frame)
        if sig == signal.SIGINT:
            sys.stdout.write("\n")
            sys.stdout.write("-- thumbor closed by user interruption --\n")

    signal.signal(signal.SIGTERM, server_sig_handler)
    signal.signal(signal.SIGINT, server_sig_handler)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from preggy import expect
from sentry_sdk import Hub
from sentry_sdk.transport import Transport

from tests.base import TestCase
from thumbor import __version__
from thumbor.config import Config
from thumbor.context import ServerParameters
from thumbor.server import get_importer


class MockTransport(Transport):
    """
    Transport that stores events in memory instead of sending them to Sentry.
    """

    def __init__(self, options=None):
        super().__init__(options=options)
        self.captured_events = []

    def capture_event(self, event):
        self.captured_events.append(event)


class FakeRequest:
    def __init__(self):
        self.headers = {
            "header1": "value1",
            "Cookie": "cookie1=value; cookie2=value2;",
        }

        self.url = "test/"
        self.method = "GET"
        self.arguments = []
        self.body = "body"
        self.query = "a=1&b=2"
        self.remote_ip = "127.0.0.1"

    def full_url(self):
        return f"http://test/{self.url}"


class FakeHandler:
    def __init__(self):
        self.request = FakeRequest()


class InvalidSentryTestCase(TestCase):
    def get_importer(self):
        return get_importer(self.config)

    def test_when_missing_dsn(self):
        self.config = Config(
            SECURITY_KEY="ACME-SEC",
            USE_CUSTOM_ERROR_HANDLING=True,
            ERROR_HANDLER_MODULE="thumbor.error_handlers.sentry",
        )
        with expect.error_to_happen(RuntimeError):
            self.get_importer()

    def test_when_invalid_path(self):
        self.config = Config(
            SECURITY_KEY="ACME-SEC",
            USE_CUSTOM_ERROR_HANDLING=True,
            SENTRY_DSN_URL="https://key@sentry.io/12345",
            ERROR_HANDLER_MODULE="thumbor.error_handlers.invalid",
        )
        with expect.error_to_happen(ImportError):
            self.get_importer()


class SentryTestCase(TestCase):
    def setUp(self):
        super().setUp()
        self.app = self.get_app()

        self.transport_mock = MockTransport()
        Hub.current.client.transport = self.transport_mock

        self.http_handler = FakeHandler()

    def get_server(self):
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        return server

    def get_config(self):
        return Config(
            SECURITY_KEY="ACME-SEC",
            SENTRY_DSN_URL="https://key@sentry.io/12345",
            USE_CUSTOM_ERROR_HANDLING=True,
            ERROR_HANDLER_MODULE="thumbor.error_handlers.sentry",
            SENTRY_ENVIRONMENT="env",
        )

    def get_importer(self):
        return get_importer(self.config)

    def test_when_error_occurs_should_have_called_client(self):
        self.importer.error_handler.handle_error(
            self.context, self.http_handler, RuntimeError("Test")
        )

        expect(self.transport_mock.captured_events).not_to_be_empty()
        expect(self.transport_mock.captured_events).to_length(1)

        event = self.transport_mock.captured_events[0]

        expect(event["exception"]["values"][0]["type"]).to_equal(
            "RuntimeError"
        )

        expect(event["environment"]).to_equal("env")

        expect(event).to_include("extra")
        expect(event).to_include("request")
        request, extra = event["request"], event["extra"]

        expect(extra).to_include("thumbor-version")
        expect(extra["thumbor-version"]).to_equal(__version__)

        expect(extra).to_include("thumbor-loader")
        expect(extra["thumbor-loader"]).to_equal(self.config.LOADER)

        expect(extra).to_include("thumbor-storage")
        expect(extra["thumbor-storage"]).to_equal(self.config.STORAGE)

        expect(request).to_include("headers")
        expect(request["headers"]).to_length(2)

        expect(request["headers"]).to_include("Cookie")
        expect(request["headers"]["Cookie"]).to_equal(
            "cookie1=value; cookie2=value2;"
        )
        expect(request["headers"]["header1"]).to_equal("value1")

        expect(request["query_string"]).to_equal("a=1&b=2")

        expect(event["modules"]).not_to_be_empty()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import preggy

POSITIONS = [
    # length (either width or height), position in percent, expected string
    (800, "-20p", "-160"),
    (800, "30p", "240"),
    (800, "230p", "1840"),
    (50, "37p", "19"),
    (55, "53p", "29"),
    (55, "-53p", "-29"),
    (800, "center", "center"),
    (800, "30", "30"),
    (800, "-40", "-40"),
    (800, "repeat", "repeat"),
]


SOURCE_IMAGE_SIZES = [
    (800, 600),
    (600, 600),
    (600, 800),
]

WATERMARK_IMAGE_SIZES = [
    # bigger ones
    (1200, 900),
    (900, 900),
    (900, 1200),
    # one size bigger
    (1200, 500),
    (700, 400),
    (400, 700),
    # equal
    (800, 600),
    (600, 600),
    # smaller
    (500, 300),
    (300, 300),
    (300, 500),
]

RATIOS = [
    # only X
    (300, None),
    (200, None),
    (100, None),
    (50, None),
    (25, None),
    # only Y
    (None, 300),
    (None, 200),
    (None, 100),
    (None, 50),
    (None, 25),
    # X and Y
    (300, 300),
    (200, 200),
    (100, 100),
    (50, 50),
    (25, 25),
    (300, 25),
    (300, 50),
    (300, 100),
    (300, 200),
    (25, 300),
    (50, 300),
    (100, 300),
    (200, 300),
    (25, 50),
    (50, 25),
]


@preggy.assertion
def to_fit_into(topic, boundary, **kwargs):
    assert (
        topic <= boundary
    ), f"Expected topic({topic}) to fit into boundary {boundary} with test: {kwargs}"


@preggy.assertion
def to_be_true_with_additional_info(topic, **kwargs):
    assert topic, f"Expected topic to be true with test: {kwargs}"


@preggy.assertion
def to_be_equal_with_additional_info(topic, expected, **kwargs):
    assert (
        topic == expected
    ), f"Expected topic({topic}) to be ({expected}) with test: {kwargs}"


@preggy.assertion
def to_almost_equal(topic, expected, differ, **kwargs):
    assert abs(1 - topic / expected) <= (differ / 100.0), (
        f"Expected topic({topic}) to be almost equal expected"
        f"({expected}) differing only in {differ}% with test: {kwargs}"
    )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import tempfile
from datetime import datetime
from os.path import abspath, dirname, join
from unittest import mock

from preggy import expect
from tornado.testing import gen_test

from thumbor.config import Config
from thumbor.context import RequestParameters
from thumbor.result_storages import ResultStorageResult
from thumbor.result_storages.file_storage import Storage as FileStorage
from thumbor.testing import TestCase


class BaseFileStorageTestCase(TestCase):
    def __init__(self, *args, **kw):
        self.storage_path = None
        self.context = None
        self.file_storage = None
        super().__init__(*args, **kw)

    def get_config(self):
        config = super().get_config()
        self.storage_path = (
            tempfile.TemporaryDirectory()  # pylint: disable=consider-using-with
        )
        config.RESULT_STORAGE_FILE_STORAGE_ROOT_PATH = self.storage_path.name
        return config

    def tearDown(self):
        super().tearDown()
        if self.storage_path is not None:
            self.storage_path.cleanup()

    @staticmethod
    def get_request():
        return RequestParameters()

    @staticmethod
    def get_fixture_path():
        return abspath(join(dirname(__file__), "../fixtures/result_storages"))

    def get_context(self):
        ctx = super().get_context()
        cfg = self.get_config()
        ctx.config = cfg
        ctx.request = self.get_request()
        self.context = ctx
        self.file_storage = FileStorage(self.context)
        return ctx

    @staticmethod
    def get_http_path():
        return "http://example.com/path/to/a.jpg"


class FileStorageTestCase(BaseFileStorageTestCase):
    @gen_test
    async def test_normalized_path(self):
        expect(self.file_storage).not_to_be_null()
        expect(
            self.file_storage.normalize_path(self.get_http_path())
        ).to_equal(
            f"{self.storage_path.name}/default/b6/be/"
            "a3e916129541a9e7146f69a15eb4d7c77c98"
        )


class WebPFileStorageTestCase(BaseFileStorageTestCase):
    def get_config(self):
        self.storage_path = (
            tempfile.TemporaryDirectory()  # pylint: disable=consider-using-with
        )
        return Config(
            AUTO_WEBP=True,
            RESULT_STORAGE_FILE_STORAGE_ROOT_PATH=self.storage_path.name,
        )

    def tearDown(self):
        super().tearDown()
        self.storage_path.cleanup()

    def get_request(self):  # pylint: disable=arguments-differ
        return RequestParameters(accepts_webp=True)

    @gen_test
    async def test_normalized_path_with_auto_webp_path(self):
        expect(self.file_storage).not_to_be_null()
        expect(
            self.file_storage.normalize_path(self.get_http_path())
        ).to_equal(
            f"{self.storage_path.name}/auto_webp/b6/be/"
            "a3e916129541a9e7146f69a15eb4d7c77c98"
        )


class ResultStorageResultTestCase(BaseFileStorageTestCase):
    def get_config(self):
        return Config(
            RESULT_STORAGE_FILE_STORAGE_ROOT_PATH=self.get_fixture_path()
        )

    def get_request(self):  # pylint: disable=arguments-differ
        return RequestParameters(url="image.jpg")

    @gen_test
    async def test_can_get_image_from_storage(self):
        result = await self.file_storage.get()

        expect(result).to_be_instance_of(ResultStorageResult)
        expect(result.successful).to_equal(True)
        expect(len(result)).to_equal(5319)
        expect(len(result)).to_equal(result.metadata["ContentLength"])
        expect(result.last_modified).to_be_instance_of(datetime)


class ExpiredFileStorageTestCase(BaseFileStorageTestCase):
    def get_config(self):
        return Config(
            RESULT_STORAGE_FILE_STORAGE_ROOT_PATH=self.get_fixture_path(),
            RESULT_STORAGE_EXPIRATION_SECONDS=10,
        )

    def get_request(self):  # pylint: disable=arguments-differ
        return RequestParameters(url="image.jpg")

    @gen_test
    async def test_cannot_get_expired_1_day_old_image(self):
        current_timestamp = (
            datetime.utcnow() - datetime(1970, 1, 1)
        ).total_seconds()
        new_mtime = current_timestamp - 60 * 60 * 24
        with mock.patch(
            "thumbor.result_storages.file_storage.getmtime",
            return_value=new_mtime,
        ):
            result = await self.file_storage.get()
        expect(result).to_be_null()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from urllib.parse import quote, unquote

from thumbor.context import RequestParameters
from thumbor.handlers import ContextHandler


class ImagingHandler(ContextHandler):
    def compute_etag(self):
        if self.context.config.ENABLE_ETAGS:
            return super().compute_etag()
        return None

    async def check_image(
        self, kwargs
    ):  # pylint: disable=too-many-return-statements
        if self.context.config.MAX_ID_LENGTH > 0:
            # Check if an image with an uuid exists in storage
            exists = await self.context.modules.storage.exists(
                kwargs["image"][: self.context.config.MAX_ID_LENGTH]
            )
            if exists:
                kwargs["image"] = kwargs["image"][
                    : self.context.config.MAX_ID_LENGTH
                ]

        url = self.request.path

        kwargs["image"] = quote(kwargs["image"].encode("utf-8"))
        if not self.validate(kwargs["image"]):
            self._error(
                400, "No original image was specified in the given URL"
            )
            return

        kwargs["request"] = self.request
        self.context.request = RequestParameters(**kwargs)

        has_none = (
            not self.context.request.unsafe and not self.context.request.hash
        )
        has_both = self.context.request.unsafe and self.context.request.hash

        if has_none or has_both:
            self._error(
                400, f"URL does not have hash or unsafe, or has both: {url}"
            )
            return

        if (
            self.context.request.unsafe
            and not self.context.config.ALLOW_UNSAFE_URL
        ):
            self._error(
                400,
                f"URL has unsafe but unsafe is not allowed by the config: {url}",
            )
            return

        if self.context.config.USE_BLACKLIST:
            blacklist = await self.get_blacklist_contents()
            if self.context.request.image_url in blacklist:
                self._error(
                    400,
                    f"Source image url has been blacklisted: {self.context.request.image_url}",
                )
                return

        url_signature = self.context.request.hash
        if url_signature:
            signer = self.context.modules.url_signer(
                self.context.server.security_key
            )

            try:
                quoted_hash = quote(self.context.request.hash)
            except KeyError:
                self._error(400, f"Invalid hash: {self.context.request.hash}")
                return

            url_to_validate = url.replace(
                f"/{self.context.request.hash}/", ""
            ).replace(f"/{quoted_hash}/", "")

            valid = signer.validate(
                unquote(url_signature).encode(), url_to_validate
            )

            if (
                not valid
                and self.context.config.STORES_CRYPTO_KEY_FOR_EACH_IMAGE
            ):
                # Retrieves security key for this image if it has been seen before
                security_key = await self.context.modules.storage.get_crypto(
                    self.context.request.image_url
                )
                if security_key is not None:
                    signer = self.context.modules.url_signer(security_key)
                    valid = signer.validate(
                        url_signature.encode(), url_to_validate
                    )

            if not valid:
                self._error(400, f"Malformed URL: {url}")
                return

        return await self.execute_image_operations()

    async def get(self, **kw):
        return await self.check_image(kw)

    async def head(self, **kw):
        return await self.check_image(kw)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from preggy import expect
from tornado.testing import gen_test

from tests.base import FilterTestCase


class FormatFilterTestCase(FilterTestCase):
    @gen_test
    async def test_invalid_format_should_be_null(self):
        await self.get_filtered(
            "source.jpg", "thumbor.filters.format", "format(invalid)"
        )
        expect(self.context.request.format).to_be_null()

    @gen_test
    async def test_can_set_proper_format(self):
        await self.get_filtered(
            "source.jpg", "thumbor.filters.format", "format(webp)"
        )
        expect(self.context.request.format).to_equal("webp")



# content of conftest.py

# pylint: disable=unused-import

import pytest

import tests.base  # NOQA


@pytest.fixture
def doctor_output():
    return """Using configuration file found at ./tests/invalid-thumbor.conf

Thumbor doctor will analyze your install and verify if everything is working as expected.

Verifying libraries support...

✅ pycurl is installed correctly.
✅ cairosvg is installed correctly.

Verifying thumbor compiled extensions...

✅ _alpha
✅ _bounding_box
✅ _brightness
✅ _colorize
✅ _composite
✅ _contrast
✅ _convolution
✅ _curve
✅ _equalize
✅ _fill
✅ _nine_patch
✅ _noise
✅ _rgb
✅ _round_corner
✅ _saturation
✅ _sharpen

Verifying extensibility modules found in your thumbor.conf...

❎ thumbor.storages.file_storagee - Storage for source images could not be imported.
❎ thumbor.loaders.http_loaderer - Loader for source images could not be imported.
❎ thumbor.result_storages.file_storagee - ResultStorage could not be imported.
❎ thumbor.engines.pillage - Engine for transforming images could not be imported.
❎ thumbor.storages.file_storager - Uploading to thumbor is enabled and the Upload Storage could not be imported.
✅ thumbor.detectors.face_detector
❎ thumbor.detectors.face_detectorer - Detector could not be imported.
❎ thumbor.detectors.other_invalid - Detector could not be imported.
❎ invalid-filter - Filter could not be imported.
✅ thumbor.filters.redeye
❎ thumbor.optimizers.gifver - Optimizer could not be imported.
❎ thumbor.optimizers.jpegtraner - Optimizer could not be imported.
❎ thumbor.error_handlers.sentryer - Custom error handling is enabled and the error handler module could not be imported.
❎ my.invalid.handler - Custom http handler could not be imported.
✅ thumbor.handler_lists.blacklist
✅ thumbor.handler_lists.healthcheck
✅ thumbor.handler_lists.upload

Verifying security...

❎ Using default security key.
❎ Allowing unsafe URLs.

😞 Oh no! We found some things that could improve... 😞

⚠️Warnings⚠️
* Security
    Error Message:
        Using default security key configuration in thumbor.conf.

    Error Description:
        You should specify a unique security key for thumbor or use a command line param to specify a security key.
\tFor more information visit https://thumbor.readthedocs.io/en/latest/running.html

⛔Errors⛔
* thumbor.storages.file_storagee
    Error Message:
        No module named 'thumbor.storages.file_storagee'

    Error Description:
        Storage for source images could not be imported.

* thumbor.loaders.http_loaderer
    Error Message:
        No module named 'thumbor.loaders.http_loaderer'

    Error Description:
        Loader for source images could not be imported.

* thumbor.result_storages.file_storagee
    Error Message:
        No module named 'thumbor.result_storages.file_storagee'

    Error Description:
        ResultStorage could not be imported.

* thumbor.engines.pillage
    Error Message:
        No module named 'thumbor.engines.pillage'

    Error Description:
        Engine for transforming images could not be imported.

* thumbor.storages.file_storager
    Error Message:
        No module named 'thumbor.storages.file_storager'

    Error Description:
        Uploading to thumbor is enabled and the Upload Storage could not be imported.

* thumbor.detectors.face_detectorer
    Error Message:
        No module named 'thumbor.detectors.face_detectorer'

    Error Description:
        Detector could not be imported.

* thumbor.detectors.other_invalid
    Error Message:
        No module named 'thumbor.detectors.other_invalid'

    Error Description:
        Detector could not be imported.

* invalid-filter
    Error Message:
        No module named 'invalid-filter'

    Error Description:
        Filter could not be imported.

* thumbor.optimizers.gifver
    Error Message:
        No module named 'thumbor.optimizers.gifver'

    Error Description:
        Optimizer could not be imported.

* thumbor.optimizers.jpegtraner
    Error Message:
        No module named 'thumbor.optimizers.jpegtraner'

    Error Description:
        Optimizer could not be imported.

* thumbor.error_handlers.sentryer
    Error Message:
        No module named 'thumbor.error_handlers.sentryer'

    Error Description:
        Custom error handling is enabled and the error handler module could not be imported.

* my.invalid.handler
    Error Message:
        No module named 'my'

    Error Description:
        Custom http handler could not be imported.

* Security
    Error Message:
        Unsafe URLs are enabled.

    Error Description:
        It is STRONGLY recommended that you turn off ALLOW_UNSAFE_URLS flag in production environments as this can lead to DDoS attacks against thumbor.
\tFor more information visit https://thumbor.readthedocs.io/en/latest/security.html

❓Need Help❓

If you don't know how to fix the above problems, please open an issue with thumbor.
Don't forget to copy this log and add it to the description.
Open an issue at https://github.com/thumbor/thumbor/issues/new
"""


@pytest.fixture
def doctor_output_no_config():
    return """
Thumbor doctor will analyze your install and verify if everything is working as expected.

Verifying libraries support...

✅ pycurl is installed correctly.
✅ cairosvg is installed correctly.

Verifying thumbor compiled extensions...

✅ _alpha
✅ _bounding_box
✅ _brightness
✅ _colorize
✅ _composite
✅ _contrast
✅ _convolution
✅ _curve
✅ _equalize
✅ _fill
✅ _nine_patch
✅ _noise
✅ _rgb
✅ _round_corner
✅ _saturation
✅ _sharpen

Verifying extension programs...

✅ jpegtran is installed correctly.
✅ ffmpeg is installed correctly.
✅ gifsicle is installed correctly.
Verifying security...


🎉 Congratulations! No errors found! 🎉
"""



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor_plugins.optimizers import gifv

from thumbor.utils import logger


class Optimizer(gifv.Optimizer):
    logger.warning(
        """
    This optimizer will be deprecated on the next major version.
    Add thumbor-plugins-gifv as a dependency, and
    change thumbor.optimizers.gifv to thumbor_plugins.optimizers.gifv
    on the OPTIMIZERS list on you thumbor.conf.
    """
    )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from unittest import TestCase

from preggy import expect

from thumbor.console import get_server_parameters


class ConsoleTestCase(TestCase):
    def test_can_get_default_server_parameters(self):
        params = get_server_parameters()
        expect(params.port).to_equal(8888)
        expect(params.ip).to_equal("0.0.0.0")
        expect(params.config_path).to_be_null()
        expect(params.keyfile).to_be_null()
        expect(params.log_level).to_equal("warning")
        expect(params.app_class).to_equal("thumbor.app.ThumborServiceApp")
        expect(params.fd).to_be_null()
        expect(params.processes).to_equal(1)

    def test_can_get_custom_server_parameters(self):
        params = get_server_parameters(
            [
                "--port=9999",
                "--ip=127.0.0.1",
                "--conf=/tmp/conf.conf",
                "--keyfile=./tests/fixtures/thumbor.key",
                "--log-level=debug",
                "--app=custom.app",
                "--fd=/tmp/fd",
                "--processes=5",
            ]
        )
        expect(params.port).to_equal(9999)
        expect(params.ip).to_equal("127.0.0.1")
        expect(params.config_path).to_equal("/tmp/conf.conf")
        expect(params.keyfile).to_equal("./tests/fixtures/thumbor.key")
        expect(params.log_level).to_equal("debug")
        expect(params.app_class).to_equal("custom.app")
        expect(params.fd).to_equal("/tmp/fd")
        expect(params.processes).to_equal(5)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com


from preggy import expect
from tornado.testing import gen_test

from tests.base import FilterTestCase
from thumbor.point import FocalPoint


class RedEyeFilterTestCase(FilterTestCase):
    @gen_test
    async def test_redeye_passthrough_when_no_face_points(self):
        image = await self.get_filtered(
            "redeye.png", "thumbor.filters.redeye", "red_eye()"
        )

        expected = self.get_fixture("redeye.png")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)

    @gen_test
    async def test_redeye_applied(self):
        expected = self.get_fixture("redeye_applied.png")

        def config_context(context):
            image_w, image_h = expected.size[0], expected.size[1]
            point = FocalPoint.from_square(
                50, 50, image_w - 100, image_h - 100, origin="Face Detection"
            )

            context.request.focal_points = [point]

        image = await self.get_filtered(
            "redeye.png", "thumbor.filters.redeye", "red_eye()", config_context
        )

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.99)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import collections
import re

BUILTIN_FILTERS = [
    "thumbor.filters.brightness",
    "thumbor.filters.colorize",
    "thumbor.filters.contrast",
    "thumbor.filters.rgb",
    "thumbor.filters.round_corner",
    "thumbor.filters.quality",
    "thumbor.filters.noise",
    "thumbor.filters.watermark",
    "thumbor.filters.equalize",
    "thumbor.filters.fill",
    "thumbor.filters.sharpen",
    "thumbor.filters.strip_exif",
    "thumbor.filters.strip_icc",
    "thumbor.filters.frame",
    "thumbor.filters.grayscale",
    "thumbor.filters.rotate",
    "thumbor.filters.format",
    "thumbor.filters.max_bytes",
    "thumbor.filters.convolution",
    "thumbor.filters.blur",
    "thumbor.filters.extract_focal",
    "thumbor.filters.focal",
    "thumbor.filters.no_upscale",
    "thumbor.filters.saturation",
    "thumbor.filters.max_age",
    "thumbor.filters.curve",
    "thumbor.filters.background_color",
    "thumbor.filters.upscale",
    "thumbor.filters.proportion",
    "thumbor.filters.stretch",
]

STRIP_QUOTE = re.compile(r"^'(.+)'$")
PHASE_POST_TRANSFORM = "post_transform"
PHASE_PRE_LOAD = "pre-load"
PHASE_AFTER_LOAD = "after-load"


def filter_method(*args):
    def _filter_deco(filtered_function):
        async def wrapper(self, *args2):
            return await filtered_function(self, *args2)

        defaults = None
        if filtered_function.__defaults__:
            default_padding = [None] * (
                len(args) - (len(filtered_function.__defaults__))
            )
            defaults = default_padding + list(filtered_function.__defaults__)

        wrapper.filter_data = {
            "name": filtered_function.__name__,
            "params": args,
            "defaults": defaults,
        }
        return wrapper

    return _filter_deco


class FiltersFactory:
    def __init__(self, filter_classes):
        self.filter_classes_map = {}

        for cls in filter_classes:
            filter_name = cls.pre_compile()
            self.filter_classes_map[filter_name] = cls

    def create_instances(self, context, filter_params):
        filter_instances = collections.defaultdict(list)
        if not filter_params:
            return FiltersRunner(filter_instances)

        filter_params = filter_params.split("):")
        last_idx = len(filter_params) - 1

        for i, param in enumerate(filter_params):
            filter_name = param.split("(")[0]
            cls = self.filter_classes_map.get(filter_name, None)

            if cls is None:
                continue

            if i != last_idx:
                param = param + ")"
            instance = cls.init_if_valid(param, context)

            if instance:
                filter_instances[
                    getattr(cls, "phase", PHASE_POST_TRANSFORM)
                ].append(instance)

        return FiltersRunner(filter_instances)


class FiltersRunner:
    def __init__(self, filter_instances):
        self.filter_instances = filter_instances

    async def apply_filters(self, phase):
        filters = self.filter_instances.get(phase, None)
        if not filters:
            return None

        while filters:
            filter_to_run = filters.pop(0)
            await filter_to_run.run()


class BaseFilter:
    PositiveNumber = {"regex": r"[\d]+", "parse": int}
    PositiveNonZeroNumber = {"regex": r"[\d]*[1-9][\d]*", "parse": int}
    NegativeNumber = {
        "regex": f"[-]{PositiveNumber['regex']}",
        "parse": int,
    }
    Number = {"regex": f"[-]?{PositiveNumber['regex']}", "parse": int}
    DecimalNumber = {
        "regex": r"[-]?(?:(?:[\d]+\.?[\d]*)|(?:[\d]*\.?[\d]+))",
        "parse": float,
    }
    Boolean = {
        "regex": r"[Tt]rue|[Ff]alse|1|0",
        "parse": lambda v: v in ("true", "True", "1"),
    }
    String = {
        "regex": r"(?:'.+?')|(?:[^,]+?)",
        "parse": lambda v: STRIP_QUOTE.sub(r"\1", v),
    }

    @classmethod
    def pre_compile(cls):
        meths = [
            f for f in list(cls.__dict__.values()) if hasattr(f, "filter_data")
        ]
        if len(meths) == 0:
            return None
        cls.runnable_method = meths[0]
        filter_data = cls.runnable_method.filter_data

        cls.compile_regex(filter_data)
        return filter_data["name"]

    @classmethod
    def compile_regex(cls, filter_data):
        params = filter_data["params"]
        defaults = filter_data.get("defaults", None)
        regexes = []
        parsers = []
        for i, param in enumerate(params):
            val = (
                (param["regex"], param["parse"])
                if (isinstance(param, dict))
                else (param, None)
            )
            comma = optional = ""
            if defaults and defaults[i] is not None:
                optional = "?"
            if i > 0:
                comma = ","
            regexes.append(f"(?:{comma}\\s*({val[0]})\\s*){optional}")
            parsers.append(val[1])

        cls.parsers = parsers
        cls.regex_str = f"{filter_data['name']}\\({''.join(regexes)}\\)"
        cls.regex = re.compile(cls.regex_str)

    @classmethod
    def init_if_valid(cls, param, context):
        instance = cls(param, context)
        if instance.params is not None:
            return instance
        return None

    def __init__(self, params, context=None):
        params = self.regex.match(params) if self.regex else None
        if params:
            params = [
                parser(param) if parser else param
                for parser, param in zip(self.parsers, params.groups())
                if param
            ]
        self.params = params
        self.context = context
        self.engine = (
            context.modules.engine if context and context.modules else None
        )

    async def run(self):
        if self.params is None:
            return

        if self.engine:
            if self.engine.is_multiple():
                engines_to_run = self.engine.frame_engines()
            else:
                engines_to_run = [self.engine]
        else:
            engines_to_run = [None]

        results = []
        for engine in engines_to_run:
            self.engine = engine
            results.append(await self.runnable_method(*self.params))

        return results



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import optparse
import sys
from io import StringIO
from unittest import TestCase, mock

from preggy import expect

import thumbor.url_composer as composer
from thumbor.url_composer import (
    get_options,
    get_parser,
    get_thumbor_params,
    main,
)


class UrlComposerTestCase(TestCase):
    @mock.patch("sys.stdout", new_callable=StringIO)
    def test_can_compose_url(self, mock_stdout):
        main(
            [
                "-k",
                "MY-SECURITY-KEY",
                "-w",
                "200",
                "-e",
                "300",
                "myserver.com/myimg.jpg",
            ]
        )

        expect(mock_stdout.getvalue()).to_equal(
            "URL:\n/G_dykuWBGyEil5JnNh9cBke0Ajo=/200x300/myserver.com/myimg.jpg\n"
        )

    @mock.patch("sys.stdout", new_callable=StringIO)
    @mock.patch.object(composer.Config, "load")
    def test_fails_when_no_config_or_key(self, load_mock, mock_stdout):
        load_mock.side_effect = RuntimeError("fail to load")

        with self.assertRaises(SystemExit) as system_error:
            main(["-w", "200", "-e", "300", "myserver.com/myimg.jpg"])

        expect(mock_stdout.getvalue()).to_equal(
            "Error: The -k or --key argument is mandatory. "
            "For more information type thumbor-url -h\n"
        )

        expect(system_error.exception.code).to_equal(1)

    def test_get_thumbor_params(self):
        params = mock.Mock(
            key_file="./tests/fixtures/thumbor.key",
            crop=None,
        )
        config = mock.Mock()

        security_key, thumbor_params = get_thumbor_params(
            "/image/url.jpg",
            params,
            config,
        )

        expect(security_key).to_equal("SECURITY_KEY_FILE")
        expect(thumbor_params["image_url"]).to_equal("/image/url.jpg")

    def test_get_thumbor_params_with_crop(self):
        params = mock.Mock(
            key_file="./tests/fixtures/thumbor.key",
            crop="300x200:400x500",
        )
        config = mock.Mock()

        security_key, thumbor_params = get_thumbor_params(
            "/image/url.jpg",
            params,
            config,
        )

        expect(security_key).to_equal("SECURITY_KEY_FILE")
        expect(thumbor_params["image_url"]).to_equal("/image/url.jpg")
        expect(thumbor_params["crop_left"]).to_equal("300")
        expect(thumbor_params["crop_top"]).to_equal("200")
        expect(thumbor_params["crop_right"]).to_equal("400")
        expect(thumbor_params["crop_bottom"]).to_equal("500")

    def test_get_thumbor_params_full_adaptive_fitin_false(self):
        params = mock.Mock(
            key_file=None,
            crop=None,
            fitin=False,
            full=False,
            adaptive=False,
        )
        config = mock.Mock(SECURITY_KEY="woot")

        security_key, thumbor_params = get_thumbor_params(
            "/image/url.jpg",
            params,
            config,
        )

        expect(security_key).not_to_be_null()
        expect(thumbor_params["fit_in"]).to_be_false()
        expect(thumbor_params["full_fit_in"]).to_be_false()
        expect(thumbor_params["adaptive_fit_in"]).to_be_false()
        expect(thumbor_params["adaptive_full_fit_in"]).to_be_false()

    def test_get_thumbor_params_full_adaptive_fitin(self):
        params = mock.Mock(
            key_file=None,
            crop=None,
            fitin=True,
            full=True,
            adaptive=True,
        )
        config = mock.Mock(SECURITY_KEY="woot")

        security_key, thumbor_params = get_thumbor_params(
            "/image/url.jpg",
            params,
            config,
        )

        expect(security_key).not_to_be_null()
        expect(thumbor_params["adaptive_full_fit_in"]).to_be_true()

    def test_get_thumbor_params_adaptive_fitin(self):
        params = mock.Mock(
            key_file=None,
            crop=None,
            fitin=True,
            full=False,
            adaptive=True,
        )
        config = mock.Mock(SECURITY_KEY="woot")

        security_key, thumbor_params = get_thumbor_params(
            "/image/url.jpg",
            params,
            config,
        )

        expect(security_key).not_to_be_null()
        expect(thumbor_params["adaptive_fit_in"]).to_be_true()

    def test_get_thumbor_params_full_fitin(self):
        params = mock.Mock(
            key_file=None,
            crop=None,
            fitin=True,
            full=True,
            adaptive=False,
        )
        config = mock.Mock(SECURITY_KEY="woot")

        security_key, thumbor_params = get_thumbor_params(
            "/image/url.jpg",
            params,
            config,
        )

        expect(security_key).not_to_be_null()
        expect(thumbor_params["full_fit_in"]).to_be_true()

    def test_get_thumbor_params_fitin(self):
        params = mock.Mock(
            key_file=None,
            crop=None,
            fitin=True,
            full=False,
            adaptive=False,
        )
        config = mock.Mock(SECURITY_KEY="woot")

        security_key, thumbor_params = get_thumbor_params(
            "/image/url.jpg",
            params,
            config,
        )

        expect(security_key).not_to_be_null()
        expect(thumbor_params["fit_in"]).to_be_true()

    def test_get_options(self):
        options, args = get_options(
            [
                "-k",
                "MY-SECURITY-KEY",
                "-w",
                "200",
                "-e",
                "300",
                "myserver.com/myimg.jpg",
            ]
        )

        expect(options.key).to_equal("MY-SECURITY-KEY")
        expect(options.width).to_equal(200)
        expect(options.height).to_equal(300)

        expect(args).to_length(1)
        expect(args[0]).to_equal("myserver.com/myimg.jpg")

    def test_get_options_from_sys(self):
        argv = [
            "thumbor-url",
            "-k",
            "MY-SECURITY-KEY",
            "-w",
            "200",
            "-e",
            "300",
            "myserver.com/myimg.jpg",
        ]
        with mock.patch.object(sys, "argv", argv):
            options, args = get_options(None)

        expect(options.key).to_equal("MY-SECURITY-KEY")
        expect(options.width).to_equal(200)
        expect(options.height).to_equal(300)

        expect(args).to_length(1)
        expect(args[0]).to_equal("myserver.com/myimg.jpg")

    def test_get_options_fails_when_no_url(self):
        parsed_options, arguments = get_options(
            ["-k", "MY-SECURITY-KEY", "-w", "200", "-e", "300"]
        )
        expect(parsed_options).to_be_null()
        expect(arguments).to_be_null()

    def test_get_parser(self):
        parser = get_parser()
        expect(parser).to_be_instance_of(optparse.OptionParser)
        expect(parser.get_usage()).to_equal(
            "Usage: thumbor-url [options] imageurl or type thumbor-url -h (--help) for help\n"
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from preggy import expect
from tornado.testing import gen_test

from tests.base import TestCase
from thumbor.context import RequestParameters
from thumbor.result_storages.no_storage import Storage as NoStorage


class NoResultStorageTestCase(TestCase):
    def get_context(self):
        ctx = super().get_context()
        ctx.request = RequestParameters(url="image.jpg")
        self.context = ctx
        return ctx

    @gen_test
    async def test_result_has_no_image(self):
        no_storage = NoStorage(self.context)
        result = await no_storage.get()
        expect(result).to_be_null()

    @gen_test
    async def test_can_put_image(self):
        no_storage = NoStorage(self.context)
        result = await no_storage.put(100)
        expect(result).to_equal("")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from preggy import expect

import thumbor.handler_lists.blacklist as handler_list
from thumbor.handlers.blacklist import BlacklistHandler
from thumbor.testing import TestCase


class BlacklistHandlerListTestCase(TestCase):
    def test_can_get_handlers(self):
        ctx = self.get_context()
        ctx.config.USE_BLACKLIST = True

        handlers = handler_list.get_handlers(ctx)

        expect(handlers).not_to_be_null()
        expect(handlers).to_length(1)
        url, handler, initializer = handlers[0]
        expect(url).to_equal(r"/blacklist/?")
        expect(handler).to_equal(BlacklistHandler)
        expect(initializer).to_equal({"context": ctx})

    def test_can_disable_blacklist(self):
        ctx = self.get_context()
        ctx.config.UPLOAD_ENABLED = False

        handlers = handler_list.get_handlers(ctx)

        expect(handlers).not_to_be_null()
        expect(handlers).to_be_empty()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com


from redis import Redis, RedisError, Sentinel
from remotecv.unique_queue import UniqueQueue

from thumbor.detectors import BaseDetector
from thumbor.utils import logger

SINGLE_NODE = "single_node"
SENTINEL = "sentinel"


class QueuedDetector(BaseDetector):
    queue = None
    detection_type = "all"

    async def detect(self):
        self.context.request.prevent_result_storage = True
        try:
            if not QueuedDetector.queue:
                redis = self.redis_client()
                QueuedDetector.queue = UniqueQueue(server=redis)

            QueuedDetector.queue.enqueue_unique_from_string(
                "remotecv.pyres_tasks.DetectTask",
                "Detect",
                args=[self.detection_type, self.context.request.image_url],
                key=self.context.request.image_url,
            )
        except RedisError:
            self.context.request.detection_error = True
            QueuedDetector.queue = None
            logger.exception("Redis Error")

        # Error or not we return an empty list as detection
        # will be done later
        return []

    def redis_client(self):
        redis_mode = str(self.context.config.REDIS_QUEUE_MODE).lower()

        if redis_mode == SINGLE_NODE:
            return self.__redis_single_node_client()
        if redis_mode == SENTINEL:
            return self.__redis_sentinel_client()

        raise RedisError(
            f"REDIS_QUEUE_MODE must be {SINGLE_NODE} or {SENTINEL}"
        )

    def __redis_single_node_client(self):
        return Redis(
            host=self.context.config.REDIS_QUEUE_SERVER_HOST,
            port=self.context.config.REDIS_QUEUE_SERVER_PORT,
            db=self.context.config.REDIS_QUEUE_SERVER_DB,
            password=self.context.config.REDIS_QUEUE_SERVER_PASSWORD,
        )

    def __redis_sentinel_client(self):
        instances_split = (
            self.context.config.REDIS_QUEUE_SENTINEL_INSTANCES.split(",")
        )
        instances = [
            tuple(instance.split(":")) for instance in instances_split
        ]

        if self.context.config.REDIS_QUEUE_SENTINEL_PASSWORD:
            sentinel_instance = Sentinel(
                instances,
                socket_timeout=self.context.config.REDIS_QUEUE_SENTINEL_SOCKET_TIMEOUT,
                sentinel_kwargs={
                    "password": self.context.config.REDIS_QUEUE_SENTINEL_PASSWORD
                },
            )
        else:
            sentinel_instance = Sentinel(
                instances,
                socket_timeout=self.context.config.REDIS_QUEUE_SENTINEL_SOCKET_TIMEOUT,
            )

        return sentinel_instance.master_for(
            self.context.config.REDIS_QUEUE_SENTINEL_MASTER_INSTANCE,
            socket_timeout=self.context.config.REDIS_QUEUE_SENTINEL_SOCKET_TIMEOUT,
            password=self.context.config.REDIS_QUEUE_SENTINEL_MASTER_PASSWORD,
            db=self.context.config.REDIS_QUEUE_SENTINEL_MASTER_DB,
        )



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import os
from io import BytesIO
from subprocess import PIPE, Popen
from tempfile import mkstemp

import piexif
from PIL import Image, ImageDraw, ImageFile, ImageSequence, JpegImagePlugin
from PIL import features as pillow_features

from thumbor.engines import BaseEngine
from thumbor.engines.extensions.pil import GifWriter
from thumbor.filters.fill import Filter
from thumbor.utils import deprecated, ensure_srgb, get_color_space, logger

try:
    from thumbor.ext.filters import _composite

    FILTERS_AVAILABLE = True
except ImportError:
    FILTERS_AVAILABLE = False

try:
    from PIL import _avif  # pylint: disable=ungrouped-imports
except ImportError:
    try:
        from pillow_avif import _avif
    except ImportError:
        _avif = None

try:
    from pillow_heif import HeifImagePlugin
except ImportError:
    HeifImagePlugin = None

HAVE_AVIF = _avif is not None
HAVE_HEIF = HeifImagePlugin is not None


FORMATS = {
    ".tif": "PNG",  # serve tif as png
    ".jpg": "JPEG",
    ".jpeg": "JPEG",
    ".gif": "GIF",
    ".png": "PNG",
    ".webp": "WEBP",
    ".avif": "AVIF",
    ".heic": "HEIF",
    ".heif": "HEIF",
}

KEEP_EXIF_COPYRIGHT_TAGS = [
    piexif.ImageIFD.Artist,
    piexif.ImageIFD.Copyright,
    piexif.ImageIFD.DateTime,
]

ImageFile.MAXBLOCK = 2**25
ImageFile.LOAD_TRUNCATED_IMAGES = True

DECOMPRESSION_BOMB_EXCEPTIONS = (Image.DecompressionBombWarning,)

if hasattr(Image, "DecompressionBombError"):
    DECOMPRESSION_BOMB_EXCEPTIONS += (Image.DecompressionBombError,)


class Engine(BaseEngine):
    def __init__(self, context):
        super().__init__(context)
        self.subsampling = None
        self.qtables = None
        self.original_mode = None
        self.exif = None

        try:
            if self.context.config.MAX_PIXELS is None or int(
                self.context.config.MAX_PIXELS
            ):
                Image.MAX_IMAGE_PIXELS = self.context.config.MAX_PIXELS
        except (AttributeError, TypeError, ValueError):  # invalid type
            logger.info(
                "MAX_PIXELS config variable set to invalid type. Has to be int on None"
            )

    def gen_image(self, size, color):
        if color == "transparent":
            color = None
        elif color == "auto":
            color = Filter.get_median_color(self.context.modules)
            color = f"#{color}"

        img = Image.new("RGBA", size, color)

        return img

    def create_image(self, buffer):
        try:
            img = Image.open(BytesIO(buffer))
        except DECOMPRESSION_BOMB_EXCEPTIONS as error:
            logger.warning("[PILEngine] create_image failed: %s", error)

            return None
        self.icc_profile = img.info.get("icc_profile")
        self.exif = img.info.get("exif")
        self.original_mode = img.mode

        if hasattr(img, "layer"):
            self.subsampling = JpegImagePlugin.get_sampling(img)
            if self.subsampling == -1:  # n/a for this file
                self.subsampling = None

        self.qtables = getattr(img, "quantization", None)

        if (
            self.context.config.ALLOW_ANIMATED_GIFS
            and self.extension == ".gif"
        ):
            frames = []

            for frame in ImageSequence.Iterator(img):
                frames.append(frame.convert("P"))
            img.seek(0)
            self.frame_count = len(frames)

            return frames

        return img

    def get_exif_copyright(self):
        exifs = piexif.load(self.image.info.get("exif"))
        copyright_exif = {}
        if exifs is None or "0th" not in exifs or exifs["0th"] is None:
            return None

        for copyright_tag in KEEP_EXIF_COPYRIGHT_TAGS:
            if copyright_tag in exifs["0th"]:
                copyright_exif[copyright_tag] = exifs["0th"][copyright_tag]

        return piexif.dump({"0th": copyright_exif})

    def get_resize_filter(self):
        config = self.context.config
        resample = (
            config.PILLOW_RESAMPLING_FILTER
            if config.PILLOW_RESAMPLING_FILTER is not None
            else "LANCZOS"
        )

        available = {
            "LANCZOS": Image.LANCZOS,
            "NEAREST": Image.NEAREST,
            "BILINEAR": Image.BILINEAR,
            "BICUBIC": Image.BICUBIC,
            "HAMMING": Image.HAMMING,
        }

        return available.get(resample.upper(), Image.LANCZOS)

    def draw_rectangle(self, x, y, width, height):
        # Nasty retry if the image is loaded for the first time and it's truncated
        try:
            draw_image = ImageDraw.Draw(self.image)
        except IOError:
            draw_image = ImageDraw.Draw(self.image)
        draw_image.rectangle([x, y, x + width, y + height])

        del draw_image

    def resize(self, width, height):
        # Indexed color modes (such as 1 and P) will be forced to use a
        # nearest neighbor resampling algorithm. So we convert them to
        # RGB(A) mode before resizing to avoid nasty scaling artifacts.

        if self.image.mode in ["1", "P"]:
            logger.debug(
                "converting image from 8-bit/1-bit palette to 32-bit RGB(A) for resize"
            )

            if self.image.mode == "1":
                target_mode = "RGB"
            else:
                # convert() figures out RGB or RGBA based on palette used
                target_mode = None
            self.image = self.image.convert(mode=target_mode)

        size = (int(width), int(height))
        # Tell image loader what target size we want (only JPG for a moment)
        self.image.draft(None, size)

        resample = self.get_resize_filter()
        self.image = self.image.resize(size, resample)

    def crop(self, left, top, right, bottom):
        self.image = self.image.crop(
            (int(left), int(top), int(right), int(bottom))
        )

    def rotate(self, degrees):
        # PIL rotates counter clockwise

        if degrees == 90:
            self.image = self.image.transpose(Image.ROTATE_90)
        elif degrees == 180:
            self.image = self.image.transpose(Image.ROTATE_180)
        elif degrees == 270:
            self.image = self.image.transpose(Image.ROTATE_270)
        else:
            self.image = self.image.rotate(degrees, expand=1)

    def flip_vertically(self):
        self.image = self.image.transpose(Image.FLIP_TOP_BOTTOM)

    def flip_horizontally(self):
        self.image = self.image.transpose(Image.FLIP_LEFT_RIGHT)

    def get_default_extension(self):
        # extension is not present => force JPEG or PNG

        if self.image.mode in ["P", "RGBA", "LA"]:
            return ".png"

        return ".jpeg"

    # TODO: Refactor this
    def read(  # noqa
        self, extension=None, quality=None
    ):  # noqa pylint: disable=too-many-statements,too-many-branches
        # returns image buffer in byte format.

        img_buffer = BytesIO()
        requested_extension = extension or self.extension

        # 1 and P mode images will be much smaller if converted back to
        # their original mode. So let's do that after resizing. Get $$.

        if (
            self.context.config.PILLOW_PRESERVE_INDEXED_MODE
            and requested_extension in [None, ".png", ".gif"]
            and self.original_mode in ["P", "1"]
            and self.original_mode != self.image.mode
        ):
            if self.original_mode == "1":
                self.image = self.image.convert("1")
            else:
                # libimagequant might not be enabled on compile time
                # but it's better than default octree for RGBA images
                quantize_method = (
                    Image.LIBIMAGEQUANT
                    if pillow_features.check("libimagequant")
                    else None
                )
                self.image = self.image.quantize(method=quantize_method)

        ext = requested_extension or self.get_default_extension()

        if ext in (".heic", ".heif") and not HAVE_HEIF:
            logger.warning(
                "[PILEngine] HEIF encoding unavailable, defaulting to %s",
                self.extension,
            )
            ext = self.extension

        if ext == ".avif" and not HAVE_AVIF:
            logger.warning(
                "[PILEngine] AVIF encoding unavailable, defaulting to %s",
                self.extension,
            )
            ext = self.extension

        options = {"quality": quality}

        if ext in (".jpg", ".jpeg"):
            options["optimize"] = True

            if self.context.config.PROGRESSIVE_JPEG:
                # Can't simply set options['progressive'] to the value
                # of self.context.config.PROGRESSIVE_JPEG because save
                # operates on the presence of the key in **options, not
                # the value of that setting.
                options["progressive"] = True

            if self.image.mode != "RGB":
                self.image = self.image.convert("RGB")
            else:
                subsampling_config = (
                    self.context.config.PILLOW_JPEG_SUBSAMPLING
                )
                qtables_config = self.context.config.PILLOW_JPEG_QTABLES

                if (
                    subsampling_config is not None
                    or qtables_config is not None
                ):
                    # can't use 'keep' here as Pillow would try to extract
                    # qtables/subsampling and fail
                    options["quality"] = 0

                    orig_subsampling = self.subsampling
                    orig_qtables = self.qtables

                    if (
                        subsampling_config == "keep"
                        or subsampling_config is None
                    ) and (orig_subsampling is not None):
                        options["subsampling"] = orig_subsampling
                    else:
                        options["subsampling"] = subsampling_config

                    if (
                        qtables_config == "keep" or qtables_config is None
                    ) and (orig_qtables and 2 <= len(orig_qtables) <= 4):
                        options["qtables"] = orig_qtables
                    else:
                        options["qtables"] = qtables_config

        if (
            ext == ".png"
            and self.context.config.PNG_COMPRESSION_LEVEL is not None
        ):
            options[
                "compress_level"
            ] = self.context.config.PNG_COMPRESSION_LEVEL

        if options["quality"] is None:
            options["quality"] = self.context.config.QUALITY

        if ext == ".avif":
            options["codec"] = self.context.config.AVIF_CODEC
            if self.context.config.AVIF_SPEED:
                options["speed"] = self.context.config.AVIF_SPEED

            if options["codec"] == "svt":
                width, height = self.size
                # SVT-AV1 has limits on min and max image dimension. If the
                # image falls outside of those, use AVIF_CODEC_FALLBACK
                if not 64 <= width <= 4096 or not 64 <= height <= 4096:
                    options["codec"] = self.context.config.AVIF_CODEC_FALLBACK
                elif width % 2 or height % 2:
                    # SVT-AV1 requires width and height to be divisible by two
                    width = (width // 2) * 2
                    height = (height // 2) * 2
                    self.crop(0, 0, width, height)

            if options["quality"] == "keep":
                options.pop("quality")

            if self.image.mode not in ["RGB", "RGBA"]:
                if self.image.mode == "P":
                    mode = "RGBA"
                else:
                    mode = "RGBA" if self.image.mode[-1] == "A" else "RGB"
                self.image = self.image.convert(mode)

            # Some AVIF decoders (most notably the one in Chrome) do not
            # display AVIF images if they have an embedded ICC profile with a
            # color space that doesn't match the image's mode (e.g. if the
            # mode is RGB but the profile is CMYK or GRAY).
            #
            # To address this issue we transform non-sRGB ICC profiles to sRGB
            # if we're encoding to AVIF.
            color_space = get_color_space(self.image)
            if color_space not in ("RGB", None):
                srgb_image = ensure_srgb(
                    self.image, srgb_profile=self.context.config.SRGB_PROFILE
                )
                if srgb_image:
                    self.image = srgb_image
                    self.icc_profile = srgb_image.info.get("icc_profile")

        if self.icc_profile is not None:
            options["icc_profile"] = self.icc_profile

        if (
            self.context.config.PRESERVE_EXIF_COPYRIGHT_INFO is True
            and self.image.info.get("exif") is not None
        ):
            exif_copyright = self.get_exif_copyright()
            if exif_copyright is not None:
                options["exif"] = exif_copyright

        if self.context.config.PRESERVE_EXIF_INFO:
            if self.exif is not None:
                options["exif"] = self.exif

        try:
            if ext == ".webp":
                if options["quality"] == 100:
                    logger.debug("webp quality is 100, using lossless instead")
                    options["lossless"] = True
                    options.pop("quality")

                if self.image.mode not in ["RGB", "RGBA"]:
                    if self.image.mode == "P":
                        mode = "RGBA"
                    else:
                        mode = "RGBA" if self.image.mode[-1] == "A" else "RGB"
                    self.image = self.image.convert(mode)

            if (
                ext in [".png", ".gif", ".heic", ".heif"]
                and self.image.mode == "CMYK"
            ):
                # 26.10.22: remove ".heic, .heif" in a month(when pillow_heif get updated)
                self.image = self.image.convert("RGBA")

            self.image.format = FORMATS.get(
                ext, FORMATS[self.get_default_extension()]
            )
            self.image.save(img_buffer, self.image.format, **options)
        except IOError:
            logger.exception(
                "Could not save as improved image, consider to increase ImageFile.MAXBLOCK"
            )
            self.image.save(img_buffer, FORMATS[ext])

        results = img_buffer.getvalue()
        img_buffer.close()
        self.extension = ext

        return results

    def read_multiple(self, images, extension=None):
        gif_writer = GifWriter()
        img_buffer = BytesIO()

        duration = []
        converted_images = []
        coordinates = []
        dispose = []

        for image in images:
            duration.append(image.info.get("duration", 80) / 1000)
            converted_images.append(image.convert("RGB"))
            coordinates.append((0, 0))
            dispose.append(1)

        loop = int(self.image.info.get("loop", 1))

        images = gif_writer.convertImagesToPIL(converted_images, False, None)
        gif_writer.writeGifToFile(
            img_buffer, images, duration, loop, coordinates, dispose
        )

        results = img_buffer.getvalue()
        img_buffer.close()

        tmp_fd, tmp_file_path = mkstemp()
        temp_file = os.fdopen(tmp_fd, "wb")
        temp_file.write(results)
        temp_file.close()

        command = ["gifsicle", "--colors", "256", tmp_file_path]

        popen = Popen(  # pylint: disable=consider-using-with
            command, stdout=PIPE
        )
        pipe = popen.stdout
        pipe_output = pipe.read()
        pipe.close()

        if popen.wait() == 0:
            results = pipe_output

        os.remove(tmp_file_path)

        return results

    @deprecated("Use image_data_as_rgb instead.")
    def get_image_data(self):
        return self.image.tobytes()

    def set_image_data(self, data):
        self.image.frombytes(data)

    @deprecated("Use image_data_as_rgb instead.")
    def get_image_mode(self):
        return self.image.mode

    def image_data_as_rgb(self, update_image=True):
        converted_image = self.image

        if converted_image.mode not in ["RGB", "RGBA"]:
            if "A" in converted_image.mode:
                converted_image = converted_image.convert("RGBA")
            elif converted_image.mode == "P":
                # convert() figures out RGB or RGBA based on palette used
                converted_image = converted_image.convert(None)
            else:
                converted_image = converted_image.convert("RGB")

        if update_image:
            self.image = converted_image

        return converted_image.mode, converted_image.tobytes()

    def convert_to_grayscale(self, update_image=True, alpha=True):
        if "A" in self.image.mode and alpha:
            image = self.image.convert("LA")
        else:
            image = self.image.convert("L")

        if update_image:
            self.image = image

        return image

    def has_transparency(self):
        has_transparency = (
            "A" in self.image.mode or "transparency" in self.image.info
        )

        if has_transparency:
            # If the image has alpha channel,
            # we check for any pixels that are not opaque (255)
            has_transparency = (
                min(self.image.convert("RGBA").getchannel("A").getextrema())
                < 255
            )

        return has_transparency

    def paste(self, other_engine, pos, merge=True):
        if merge and not FILTERS_AVAILABLE:
            raise RuntimeError(
                "You need filters enabled to use paste with merge. Please reinstall "
                + "thumbor with proper compilation of its filters."
            )

        self.enable_alpha()
        other_engine.enable_alpha()

        image = self.image
        other_image = other_engine.image

        if merge:
            image_size = self.size
            other_size = other_engine.size
            mode, data = self.image_data_as_rgb()
            _, other_data = other_engine.image_data_as_rgb()
            imgdata = _composite.apply(
                mode,
                data,
                image_size[0],
                image_size[1],
                other_data,
                other_size[0],
                other_size[1],
                int(pos[0]),
                int(pos[1]),
            )
            self.set_image_data(imgdata)
        else:
            image.paste(other_image, pos)

    def enable_alpha(self):
        if self.image.mode != "RGBA":
            self.image = self.image.convert("RGBA")

    def strip_icc(self):
        self.icc_profile = None

    def strip_exif(self):
        self.exif = None



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com
from unittest import TestCase, mock

from preggy import expect

import thumbor.server
from tests.fixtures.custom_error_handler import (
    ErrorHandler as CustomErrorHandler,
)
from thumbor.app import ThumborServiceApp
from thumbor.config import Config
from thumbor.server import (
    configure_log,
    get_application,
    get_as_integer,
    get_config,
    get_context,
    get_importer,
    main,
    run_server,
    validate_config,
)


class ServerTestCase(TestCase):
    def test_can_get_value_as_integer(self):
        expect(get_as_integer("1")).to_equal(1)
        expect(get_as_integer("a")).to_be_null()
        expect(get_as_integer("")).to_be_null()
        expect(get_as_integer(None)).to_be_null()

    def test_can_get_config_from_path(self):
        config = get_config("./tests/fixtures/thumbor_config_server_test.conf")

        with mock.patch.dict("os.environ", {"ENGINE": "test"}):
            expect(config).not_to_be_null()
            expect(config.ALLOWED_SOURCES).to_be_like(["mydomain.com"])
            expect(config.ENGINE).to_be_like("thumbor.engines.pil")

    def test_can_get_config_with_env_enabled(self):
        config = get_config(
            "./tests/fixtures/thumbor_config_server_test.conf", True
        )

        with mock.patch.dict("os.environ", {"ENGINE": "test"}):
            expect(config).not_to_be_null()
            expect(config.ALLOWED_SOURCES).to_be_like(["mydomain.com"])
            expect(config.ENGINE).to_be_like("test")

    @mock.patch("logging.basicConfig")
    def test_can_configure_log_from_config(self, basic_config_mock):
        conf = Config()
        configure_log(conf, "DEBUG")

        params = dict(
            datefmt="%Y-%m-%d %H:%M:%S",
            level=10,
            format="%(asctime)s %(name)s:%(levelname)s %(message)s",
        )

        basic_config_mock.assert_called_with(**params)

    @mock.patch("logging.config.dictConfig")
    def test_can_configure_log_from_dict_config(self, dict_config_mock):
        conf = Config(THUMBOR_LOG_CONFIG={"level": "INFO"})
        configure_log(conf, "DEBUG")

        params = dict(
            level="INFO",
        )

        dict_config_mock.assert_called_with(params)

    def test_can_import_default_modules(self):
        conf = Config()
        importer = get_importer(conf)

        expect(importer).not_to_be_null()
        expect(importer.filters).not_to_be_empty()

    def test_can_import_with_custom_error_handler_class(self):
        conf = Config(
            USE_CUSTOM_ERROR_HANDLING=True,
            ERROR_HANDLER_MODULE="tests.fixtures.custom_error_handler",
        )
        importer = get_importer(conf)

        expect(importer).not_to_be_null()
        expect(importer.error_handler_class).not_to_be_null()
        expect(importer.error_handler_class).to_be_instance_of(
            CustomErrorHandler
        )

    def test_validate_config_security_key(self):
        server_parameters = mock.Mock(security_key=None)
        conf = Config(SECURITY_KEY=None)

        with expect.error_to_happen(
            RuntimeError,
            message="No security key was found for this instance of thumbor. "
            "Please provide one using the conf file or a security key file.",
        ):
            validate_config(conf, server_parameters)

    def test_validate_config_security_key_from_config(self):
        server_parameters = mock.Mock(security_key=None)
        conf = Config(SECURITY_KEY="something")

        validate_config(conf, server_parameters)
        expect(server_parameters.security_key).to_equal("something")

    @mock.patch.object(thumbor.server, "which")
    def test_validate_gifsicle_path(self, which_mock):
        server_parameters = mock.Mock(security_key=None)
        conf = Config(SECURITY_KEY="test", USE_GIFSICLE_ENGINE=True)

        which_mock.return_value = "/usr/bin/gifsicle"

        validate_config(conf, server_parameters)
        expect(server_parameters.gifsicle_path).to_equal("/usr/bin/gifsicle")

    @mock.patch.object(thumbor.server, "which")
    def test_validate_null_gifsicle_path(self, which_mock):
        server_parameters = mock.Mock(security_key=None)
        conf = Config(SECURITY_KEY="test", USE_GIFSICLE_ENGINE=True)

        which_mock.return_value = None

        with expect.error_to_happen(
            RuntimeError,
            message="If using USE_GIFSICLE_ENGINE configuration to True, "
            "the `gifsicle` binary must be in the PATH and must be an executable.",
        ):
            validate_config(conf, server_parameters)

    def test_get_context(self):
        server_parameters = mock.Mock(
            security_key=None, app_class="thumbor.app.ThumborServiceApp"
        )
        conf = Config(SECURITY_KEY="test")
        importer = get_importer(conf)
        context = get_context(server_parameters, conf, importer)

        expect(context).not_to_be_null()

    def test_get_application(self):
        server_parameters = mock.Mock(
            security_key=None, app_class="thumbor.app.ThumborServiceApp"
        )
        conf = Config(SECURITY_KEY="test")
        importer = get_importer(conf)
        context = get_context(server_parameters, conf, importer)
        app = get_application(context)

        expect(app).not_to_be_null()
        expect(app).to_be_instance_of(ThumborServiceApp)

    @mock.patch.object(thumbor.server, "HTTPServer")
    def test_can_run_server_with_default_params(self, server_mock):
        application = mock.Mock()
        context = mock.Mock()
        context.server = mock.Mock(
            fd=None, port=1234, ip="0.0.0.0", processes=1
        )

        server_instance_mock = mock.Mock()
        server_mock.return_value = server_instance_mock

        run_server(application, context)

        server_instance_mock.bind.assert_called_with(1234, "0.0.0.0")
        server_instance_mock.start.assert_called_with(1)

    @mock.patch.object(thumbor.server, "HTTPServer")
    def test_can_run_server_with_multiple_processes(self, server_mock):
        application = mock.Mock()
        context = mock.Mock()
        context.server = mock.Mock(
            fd=None, port=1234, ip="0.0.0.0", processes=5
        )

        server_instance_mock = mock.Mock()
        server_mock.return_value = server_instance_mock

        run_server(application, context)

        server_instance_mock.start.assert_called_with(5)

    @mock.patch.object(thumbor.server, "HTTPServer")
    @mock.patch.object(thumbor.server, "socket", autospec=True)
    def test_can_run_server_with_fd(self, socket_mock, server_mock):
        application = mock.Mock()
        context = mock.Mock()
        context.server = mock.Mock(fd=11, port=1234, ip="0.0.0.0", processes=1)

        server_instance_mock = mock.Mock()
        server_mock.return_value = server_instance_mock

        run_server(application, context)
        socket_mock.assert_called_with(fileno=11)
        server_instance_mock.add_socket.assert_called_with(
            socket_mock.return_value
        )
        server_instance_mock.start.assert_called_with(1)

    @mock.patch.object(thumbor.server, "HTTPServer")
    @mock.patch.object(thumbor.server, "socket", autospec=True)
    def test_can_run_server_with_fd_non_blocking(
        self, socket_mock, server_mock
    ):
        server_parameters = mock.Mock(security_key=None)
        conf = Config(SECURITY_KEY="test", NON_BLOCKING_SOCKETS=True)
        importer = get_importer(conf)
        context = get_context(server_parameters, conf, importer)

        context.server = mock.Mock(fd=11, port=1234, ip="0.0.0.0", processes=1)

        server_instance_mock = mock.Mock()
        server_mock.return_value = server_instance_mock

        application = mock.Mock()
        run_server(application, context)
        socket_mock.assert_called_with(fileno=11)
        socket_mock.return_value.setblocking.assert_called_with(False)
        server_instance_mock.add_socket.assert_called_with(
            socket_mock.return_value
        )
        server_instance_mock.start.assert_called_with(1)

    @mock.patch.object(thumbor.server, "HTTPServer")
    @mock.patch.object(thumbor.server, "bind_unix_socket")
    def test_can_run_server_with_unix_socket(
        self, bind_unix_socket, server_mock
    ):
        application = mock.Mock()
        context = mock.Mock()
        context.server = mock.Mock(
            fd="/path/bin", port=1234, ip="0.0.0.0", processes=1
        )

        server_instance_mock = mock.Mock()
        server_mock.return_value = server_instance_mock

        bind_unix_socket.return_value = "socket mock"

        run_server(application, context)

        bind_unix_socket.assert_called_with("/path/bin")
        server_instance_mock.add_socket.assert_called_with("socket mock")
        server_instance_mock.start.assert_called_with(1)

    @mock.patch.object(thumbor.server, "HTTPServer")
    def test_run_server_returns_server(self, server_mock):
        application = mock.Mock()
        context = mock.Mock()
        context.server = mock.Mock(fd=None, port=1234, ip="0.0.0.0")

        server_instance_mock = mock.Mock()
        server_mock.return_value = server_instance_mock

        server = run_server(application, context)

        self.assertEqual(server, server_instance_mock)

    @mock.patch.object(thumbor.server, "setup_signal_handler")
    @mock.patch.object(thumbor.server, "HTTPServer")
    @mock.patch.object(thumbor.server, "get_server_parameters")
    @mock.patch("tornado.ioloop.IOLoop.instance", create=True)
    def test_can_run_main(
        self,
        ioloop_mock,
        get_server_parameters_mock,
        server_mock,
        setup_signal_handler_mock,
    ):
        server_parameters = mock.Mock(
            config_path="./tests/fixtures/thumbor_config_server_test.conf",
            log_level="DEBUG",
            debug=False,
            security_key="sec",
            app_class="thumbor.app.ThumborServiceApp",
            fd=None,
            ip="0.0.0.0",
            port=1234,
        )
        get_server_parameters_mock.return_value = server_parameters

        ioloop_instance_mock = mock.Mock()
        ioloop_mock.return_value = ioloop_instance_mock
        main()
        ioloop_instance_mock.start.assert_any_call()
        self.assertTrue(setup_signal_handler_mock.called)
        self.assertTrue(server_mock.called)

    def cleanup(self):
        ServerTestCase.cleanup_called = True



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from thumbor.ext.filters import _round_corner
from thumbor.filters import BaseFilter, filter_method


class Filter(BaseFilter):
    @filter_method(
        r"[\d]+(?:\|[\d]+)?",
        BaseFilter.PositiveNumber,
        BaseFilter.PositiveNumber,
        BaseFilter.PositiveNumber,
        BaseFilter.Boolean,
    )
    async def round_corner(self, radius, red, green, blue, transparent=False):
        width, height = self.engine.size
        radius_parts = radius.split("|")
        a_radius = int(radius_parts[0])
        b_radius = int(radius_parts[1]) if len(radius_parts) > 1 else a_radius

        if transparent:
            self.engine.enable_alpha()

        mode, data = self.engine.image_data_as_rgb()
        imgdata = _round_corner.apply(
            1,
            mode,
            a_radius,
            b_radius,
            red,
            green,
            blue,
            width,
            height,
            data,
            transparent,
        )
        self.engine.set_image_data(imgdata)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from preggy import expect
from tornado.testing import gen_test

from tests.handlers.test_base_handler import BaseImagingTestCase
from thumbor.config import Config
from thumbor.context import Context, ServerParameters
from thumbor.importer import Importer
from thumbor.loaders import LoaderResult

# pylint: disable=broad-except,abstract-method,attribute-defined-outside-init,line-too-long,too-many-public-methods
# pylint: disable=too-many-lines


class LoaderErrorTestCase(BaseImagingTestCase):
    def get_context(self):
        cfg = Config(SECURITY_KEY="ACME-SEC")
        cfg.LOADER = "thumbor.loaders.file_loader"
        cfg.FILE_LOADER_ROOT_PATH = self.loader_path
        cfg.STORAGE = "thumbor.storages.file_storage"
        cfg.FILE_STORAGE_ROOT_PATH = self.root_path

        importer = Importer(cfg)
        importer.import_modules()
        server = ServerParameters(
            8889, "localhost", "thumbor.conf", None, "info", None
        )
        server.security_key = "ACME-SEC"
        return Context(server, cfg, importer)

    @gen_test
    async def test_should_propagate_custom_loader_error(self):
        old_load = self.context.modules.loader.load

        async def load_override(*_):
            result = LoaderResult()
            result.successful = False
            result.error = 409
            return result

        self.context.modules.loader.load = load_override

        response = await self.async_fetch("/unsafe/image.jpg")

        self.context.modules.loader.load = old_load

        expect(response.code).to_equal(409)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com


from preggy import expect
from tornado.testing import gen_test

from tests.base import FilterTestCase


class NoiseFilterTestCase(FilterTestCase):
    @gen_test
    async def test_noise_filter(self):
        image = await self.get_filtered(
            "source.jpg", "thumbor.filters.noise", "noise(200,123)"
        )
        expected = self.get_fixture("noise.png")

        ssim = self.get_ssim(image, expected)
        expect(ssim).to_be_greater_than(0.98)



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

from preggy import expect
from tornado.testing import gen_test

from tests.base import TestCase
from thumbor.config import Config
from thumbor.importer import Importer


class BaseMaxAgeFilterTestCase(TestCase):
    def get_fixture_path(self, name):
        return f"./tests/fixtures/{name}"

    def get_config(self):
        return Config.load(self.get_fixture_path("max_age_conf.py"))

    def get_importer(self):
        importer = Importer(self.config)
        importer.import_modules()
        return importer


class MaxAgeFilterTestCase(BaseMaxAgeFilterTestCase):
    @gen_test
    async def test_max_age_filter_with_regular_image(self):
        response = await self.async_fetch(
            "/unsafe/smart/image.jpg", method="GET"
        )
        expect(response.code).to_equal(200)
        expect(response.headers["Cache-Control"]).to_equal("max-age=2,public")
        expect(response.headers).to_include("Expires")

    @gen_test
    async def test_max_age_url(self):
        response = await self.async_fetch(
            "/unsafe/filters:max_age(30)/image.jpg", method="GET"
        )
        expect(response.code).to_equal(200)
        expect(response.headers["Cache-Control"]).to_equal("max-age=30,public")
        expect(response.headers).to_include("Expires")


class MaxAgeDetectorFilterTestCase(BaseMaxAgeFilterTestCase):
    def get_config(self):
        config = super().get_config()
        config.DETECTORS = ["tests.fixtures.prevent_result_storage_detector"]
        return config

    @gen_test
    async def test_max_age_filter_with_non_storaged_image(self):
        response = await self.async_fetch(
            "/unsafe/smart/image.jpg", method="GET"
        )
        expect(response.code).to_equal(200)
        expect(response.headers["Cache-Control"]).to_equal("max-age=1,public")
        expect(response.headers).to_include("Expires")


class MaxAgeErrorDectectorFilterTestCase(BaseMaxAgeFilterTestCase):
    def get_config(self):
        config = super().get_config()
        config.DETECTORS = ["tests.fixtures.detection_error_detector"]
        return config

    @gen_test
    async def test_with_detection_error_image(self):
        response = await self.async_fetch(
            "/unsafe/smart/image.jpg", method="GET"
        )
        expect(response.code).to_equal(200)
        expect(response.headers["Cache-Control"]).to_equal("max-age=1,public")
        expect(response.headers).to_include("Expires")



#!/usr/bin/python
# -*- coding: utf-8 -*-

# thumbor imaging service
# https://github.com/thumbor/thumbor/wiki

# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license
# Copyright (c) 2011 globo.com thumbor@googlegroups.com

import hashlib
from datetime import datetime
from os.path import abspath, dirname, exists, getmtime, isdir, isfile, join
from shutil import move
from urllib.parse import unquote
from uuid import uuid4

import pytz

from thumbor.engines import BaseEngine
from thumbor.result_storages import BaseStorage, ResultStorageResult
from thumbor.utils import deprecated, logger


class Storage(BaseStorage):
    PATH_FORMAT_VERSION = "v2"

    @property
    def is_auto_webp(self):
        return (
            self.context.config.AUTO_WEBP and self.context.request.accepts_webp
        )

    async def put(self, image_bytes):
        file_abspath = self.normalize_path(self.context.request.url)
        if not self.validate_path(file_abspath):
            logger.warning(
                "[RESULT_STORAGE] unable to write outside root path: %s",
                file_abspath,
            )
            return
        temp_abspath = f"{file_abspath}.{str(uuid4()).replace('-', '')}"
        file_dir_abspath = dirname(file_abspath)
        logger.debug(
            "[RESULT_STORAGE] putting at %s (%s)",
            file_abspath,
            file_dir_abspath,
        )

        self.ensure_dir(file_dir_abspath)

        with open(temp_abspath, "wb") as _file:
            _file.write(image_bytes)

        move(temp_abspath, file_abspath)

    async def get(self):
        path = self.context.request.url
        file_abspath = self.normalize_path(path)

        if not self.validate_path(file_abspath):
            logger.warning(
                "[RESULT_STORAGE] unable to read from outside root path: %s",
                file_abspath,
            )
            return None

        logger.debug("[RESULT_STORAGE] getting from %s", file_abspath)

        if isdir(file_abspath):
            logger.warning(
                "[RESULT_STORAGE] cache location is a directory: %s",
                file_abspath,
            )
            return None

        if not exists(file_abspath):
            legacy_path = self.normalize_path_legacy(path)
            if isfile(legacy_path):
                logger.debug(
                    "[RESULT_STORAGE] migrating image from old location at %s",
                    legacy_path,
                )
                self.ensure_dir(dirname(file_abspath))
                move(legacy_path, file_abspath)
            else:
                logger.debug(
                    "[RESULT_STORAGE] image not found at %s", file_abspath
                )
                return None

        if self.is_expired(file_abspath):
            logger.debug("[RESULT_STORAGE] cached image has expired")
            return None

        with open(file_abspath, "rb") as image_file:
            buffer = image_file.read()

        result = ResultStorageResult(
            buffer=buffer,
            metadata={
                "LastModified": datetime.fromtimestamp(
                    getmtime(file_abspath)
                ).replace(tzinfo=pytz.utc),
                "ContentLength": len(buffer),
                "ContentType": BaseEngine.get_mimetype(buffer),
            },
        )

        return result

    def validate_path(self, path):
        return abspath(path).startswith(
            self.context.config.RESULT_STORAGE_FILE_STORAGE_ROOT_PATH
        )

    def normalize_path(self, path):
        digest = hashlib.sha1(unquote(path).encode("utf-8")).hexdigest()
        root_path = (
            self.context.config.RESULT_STORAGE_FILE_STORAGE_ROOT_PATH.rstrip(
                "/"
            )
        )
        prefix = "auto_webp" if self.is_auto_webp else "default"

        return f"{root_path}/{prefix}/{digest[:2]}/{digest[2:4]}/{digest[4:]}"

    def normalize_path_legacy(self, path):
        path = unquote(path)
        path_segments = [
            self.context.config.RESULT_STORAGE_FILE_STORAGE_ROOT_PATH.rstrip(
                "/"
            ),
            Storage.PATH_FORMAT_VERSION,
        ]
        if self.is_auto_webp:
            path_segments.append("webp")

        path_segments.extend([self.partition(path), path.lstrip("/")])

        normalized_path = join(*path_segments).replace("http://", "")
        return normalized_path

    def partition(self, path_raw):
        path = path_raw.lstrip("/")
        return join("".join(path[0:2]), "".join(path[2:4]))

    def is_expired(self, path):
        expire_in_seconds = self.context.config.get(
            "RESULT_STORAGE_EXPIRATION_SECONDS", None
        )

        if expire_in_seconds is None or expire_in_seconds == 0:
            return False

        timediff = datetime.now() - datetime.fromtimestamp(getmtime(path))
        return timediff.total_seconds() > expire_in_seconds

    @deprecated("Use result's last_modified instead")
    def last_updated(self):
        path = self.context.request.url
        file_abspath = self.normalize_path(path)
        if not self.validate_path(file_abspath):
            logger.warning(
                "[RESULT_STORAGE] unable to read from outside root path: %s",
                file_abspath,
            )
            return True
        logger.debug("[RESULT_STORAGE] getting from %s", file_abspath)

        if not exists(file_abspath) or self.is_expired(file_abspath):
            logger.debug(
                "[RESULT_STORAGE] image not found at %s", file_abspath
            )
            return True

        return datetime.fromtimestamp(getmtime(file_abspath)).replace(
            tzinfo=pytz.utc
        )
